
import threading
import sys

###############################################################################
# Server Adapter ###############################################################
###############################################################################


class ServerAdapter(object):
    quiet = False

    def __init__(self, host='127.0.0.1', port=8080, **options):
        self.options = options
        self.host = host
        self.port = int(port)

    def run(self, handler):
        pass

    @property
    def is_secure(self):
        return self.options.get('keyfile') is not None

    def __repr__(self):
        args = ', '.join(f'{k}={repr(v)}' for k, v in self.options.items())
        return f"{self.__class__}({args})"


class CGIServer(ServerAdapter):
    quiet = True

    def run(self, handler):  # pragma: no cover
        from wsgiref.handlers import CGIHandler

        def fixed_environ(environ, start_response):
            environ.setdefault('PATH_INFO', '')
            return handler(environ, start_response)
        CGIHandler().run(fixed_environ)


class FlupFCGIServer(ServerAdapter):
    def run(self, handler):  # pragma: no cover
        import flup.server.fcgi
        self.options.setdefault('bindAddress', (self.host, self.port))
        flup.server.fcgi.WSGIServer(handler, **self.options).run()


class WSGIRefServer(ServerAdapter):
    def run(self, app):  # pragma: no cover
        from wsgiref.simple_server import WSGIRequestHandler, WSGIServer
        from wsgiref.simple_server import make_server
        import socket

        class FixedHandler(WSGIRequestHandler):
            def address_string(self):  # Prevent reverse DNS lookups please.
                return self.client_address[0]
            def log_request(*args, **kw):
                if not self.quiet:
                    return WSGIRequestHandler.log_request(*args, **kw)

        handler_cls = self.options.get('handler_class', FixedHandler)
        server_cls = self.options.get('server_class', WSGIServer)

        if ':' in self.host:  # Fix wsgiref for IPv6 addresses.
            if getattr(server_cls, 'address_family') == socket.AF_INET:
                class server_cls(server_cls):
                    address_family = socket.AF_INET6

        srv = make_server(self.host, self.port, app, server_cls, handler_cls)
        srv.serve_forever()


class CherryPyServer(ServerAdapter):
    def run(self, handler): # pragma: no cover
        from cherrypy import wsgiserver
        self.options['bind_addr'] = (self.host, self.port)
        self.options['wsgi_app'] = handler

        certfile = self.options.get('certfile')
        if certfile:
            del self.options['certfile']
        keyfile = self.options.get('keyfile')
        if keyfile:
            del self.options['keyfile']

        server = wsgiserver.CherryPyWSGIServer(**self.options)
        if certfile:
            server.ssl_certificate = certfile
        if keyfile:
            server.ssl_private_key = keyfile

        try:
            server.start()
        finally:
            server.stop()


class WaitressServer(ServerAdapter):
    def run(self, handler):
        from waitress import serve
        serve(handler, host=self.host, port=self.port)


class PasteServer(ServerAdapter):
    def run(self, handler): # pragma: no cover
        from paste import httpserver
        from paste.translogger import TransLogger
        handler = TransLogger(handler, setup_console_handler=(not self.quiet))
        httpserver.serve(handler, host=self.host, port=str(self.port),
                         **self.options)


class MeinheldServer(ServerAdapter):
    def run(self, handler):
        from meinheld import server
        server.listen((self.host, self.port))
        server.run(handler)


class FapwsServer(ServerAdapter):
    """ Extremely fast webserver using libev. See http://www.fapws.org/ """
    def run(self, handler):  # pragma: no cover
        import fapws._evwsgi as evwsgi
        from fapws import base, config
        port = self.port
        if float(config.SERVER_IDENT[-2:]) > 0.4:
            # fapws3 silently changed its API in 0.5
            port = str(port)
        evwsgi.start(self.host, port)
        evwsgi.set_base_module(base)

        def app(environ, start_response):
            environ['wsgi.multiprocess'] = False
            return handler(environ, start_response)
        evwsgi.wsgi_cb(('', app))
        evwsgi.run()


class TornadoServer(ServerAdapter):
    """ The super hyped asynchronous server by facebook. Untested. """
    def run(self, handler):  # pragma: no cover
        import tornado.wsgi, tornado.httpserver, tornado.ioloop
        container = tornado.wsgi.WSGIContainer(handler)
        server = tornado.httpserver.HTTPServer(container)
        server.listen(port=self.port, address=self.host)
        tornado.ioloop.IOLoop.instance().start()


class AppEngineServer(ServerAdapter):
    """ Adapter for Google App Engine. """
    quiet = True
    def run(self, handler):
        from google.appengine.ext.webapp import util
        # A main() function in the handler script enables 'App Caching'.
        # Lets makes sure it is there. This _really_ improves performance.
        module = sys.modules.get('__main__')
        if module and not hasattr(module, 'main'):
            module.main = lambda: util.run_wsgi_app(handler)
        util.run_wsgi_app(handler)


class TwistedServer(ServerAdapter):
    """ Untested. """
    def run(self, handler):
        from twisted.web import server, wsgi
        from twisted.python.threadpool import ThreadPool
        from twisted.internet import reactor
        thread_pool = ThreadPool()
        thread_pool.start()
        reactor.addSystemEventTrigger('after', 'shutdown', thread_pool.stop)
        factory = server.Site(wsgi.WSGIResource(reactor, thread_pool, handler))
        reactor.listenTCP(self.port, factory, interface=self.host)
        reactor.run()


class DieselServer(ServerAdapter):
    """ Untested. """
    def run(self, handler):
        from diesel.protocols.wsgi import WSGIApplication
        app = WSGIApplication(handler, port=self.port)
        app.run()


class GeventServer(ServerAdapter):
    """ Untested. Options:

        * `fast` (default: False) uses libevent's http server, but has some
          issues: No streaming, no pipelining, no SSL.
        * See gevent.wsgi.WSGIServer() documentation for more options.
    """
    def run(self, handler):
        from gevent import pywsgi, local
        if not isinstance(threading.local(), local.local):
            msg = "Ombott requires gevent.monkey.patch_all() (before import)"
            raise RuntimeError(msg)
        if self.quiet:
            self.options['log'] = None
        address = (self.host, self.port)
        server = pywsgi.WSGIServer(address, handler, **self.options)
        server.serve_forever()


class GeventSocketIOServer(ServerAdapter):
    def run(self, handler):
        from socketio import server
        address = (self.host, self.port)
        server.SocketIOServer(address, handler, **self.options).serve_forever()


class GunicornServer(ServerAdapter):
    """ Untested. See http://gunicorn.org/configure.html for options. """
    def run(self, handler):
        from gunicorn.app.base import Application

        config = {'bind': "%s:%d" % (self.host, int(self.port))}
        config.update(self.options)

        class GunicornApplication(Application):
            def init(self, parser, opts, args):
                return config

            def load(self):
                return handler

        GunicornApplication().run()


class EventletServer(ServerAdapter):
    """ Untested """
    def run(self, handler):
        from eventlet import wsgi, listen
        try:
            wsgi.server(listen((self.host, self.port)), handler,
                        log_output=(not self.quiet))
        except TypeError:
            # Fallback, if we have old version of eventlet
            wsgi.server(listen((self.host, self.port)), handler)


class RocketServer(ServerAdapter):
    def run(self, handler):
        from rocket3 import Rocket3
        ssl_args = []
        if self.is_secure:
            ssl_args = [self.options['keyfile'], self.options['certfile']]
        server = Rocket3(
            (self.host, self.port, *ssl_args), 'wsgi', {'wsgi_app': handler}
        )
        server.start()


class BjoernServer(ServerAdapter):
    """ Fast server written in C: https://github.com/jonashaag/bjoern """
    def run(self, handler):
        from bjoern import run
        run(handler, self.host, self.port)


class AutoServer(ServerAdapter):
    """ Untested. """
    adapters = [WaitressServer, PasteServer, TwistedServer, CherryPyServer, WSGIRefServer]
    def run(self, handler):
        for sa in self.adapters:
            try:
                return sa(self.host, self.port, **self.options).run(handler)
            except ImportError:
                pass


server_names = {
    'cgi': CGIServer,
    'flup': FlupFCGIServer,
    'wsgiref': WSGIRefServer,
    'waitress': WaitressServer,
    'cherrypy': CherryPyServer,
    'paste': PasteServer,
    'fapws3': FapwsServer,
    'tornado': TornadoServer,
    'gae': AppEngineServer,
    'twisted': TwistedServer,
    'diesel': DieselServer,
    'meinheld': MeinheldServer,
    'gunicorn': GunicornServer,
    'eventlet': EventletServer,
    'gevent': GeventServer,
    'geventSocketIO': GeventSocketIOServer,
    'rocket': RocketServer,
    'bjoern': BjoernServer,
    'auto': AutoServer,
}



import os
import mimetypes
import time

import email.utils

from .common_helpers import parse_date
from .response import HTTPResponse, HTTPError
from .ombott import Globals


def get_first_range(header, maxlen=0):
    ''' Return (start, end) of first range in ranges parsed from a HTTP Range header.
        Skip unsatisfiable ranges. The end index is non-inclusive.

    '''
    try:
        ranges_str = header.split('bytes=', 1)[1]
    except IndexError:
        # no-perix/unsupported-units
        return

    first_range = ranges_str.split(',', 1)[0]
    try:
        start, end = first_range.split('-')
    except ValueError:
        # too-many/not-enougth values to unpack (invalid range)
        return

    try:
        if not start:  # bytes=-100    -> last 100 bytes
            start, end = max(0, maxlen - int(end)), maxlen
        elif not end:  # bytes=100-    -> all but the first 99 bytes
            start, end = int(start), maxlen
        else:          # bytes=100-200 -> bytes 100-200 (inclusive)
            start, end = int(start), min(int(end) + 1, maxlen)
        if 0 <= start < end <= maxlen:
            return start, end
    except ValueError:
        pass


def _file_iter_range(fp, offset, bytes_len, maxread=1024 * 1024):
    ''' Yield chunks from a range in a file. No chunk is bigger than maxread.'''
    fp.seek(offset)
    part = fp.read(min(bytes_len, maxread))
    while bytes_len > 0 and part:
        bytes_len -= len(part)
        yield part
        part = fp.read(min(bytes_len, maxread))


def static_file(filename, root, mimetype='auto', download=False, charset='UTF-8'):
    """ Open a file in a safe way and return :exc:`HTTPResponse` with status
        code 200, 305, 403 or 404. The ``Content-Type``, ``Content-Encoding``,
        ``Content-Length`` and ``Last-Modified`` headers are set if possible.
        Special support for ``If-Modified-Since``, ``Range`` and ``HEAD``
        requests.

        :param filename: Name or path of the file to send.
        :param root: Root path for file lookups. Should be an absolute directory
            path.
        :param mimetype: Defines the content-type header (default: guess from
            file extension)
        :param download: If True, ask the browser to open a `Save as...` dialog
            instead of opening the file with the associated program. You can
            specify a custom filename as a string. If not specified, the
            original filename is used (default: False).
        :param charset: The charset to use for files with a ``text/*``
            mime-type. (default: UTF-8)
    """
    request = Globals.request
    os_path = os.path
    root = os_path.abspath(root) + os.sep
    filename = os_path.abspath(os_path.join(root, filename.strip('/\\')))
    headers = dict()
    env_get = request.environ.get

    if not filename.startswith(root):
        return HTTPError(403, "Access denied.")
    if not os_path.exists(filename) or not os_path.isfile(filename):
        return HTTPError(404, "File does not exist.")
    if not os.access(filename, os.R_OK):
        return HTTPError(403, "You do not have permission to access this file.")

    if mimetype == 'auto':
        mimetype, encoding = mimetypes.guess_type(filename)
        if encoding:
            headers['Content-Encoding'] = encoding

    if mimetype:
        if mimetype.startswith('text/') and charset and 'charset' not in mimetype:
            mimetype += '; charset=%s' % charset
        headers['Content-Type'] = mimetype

    if download:
        download = os_path.basename(filename if download is True else download)
        headers['Content-Disposition'] = 'attachment; filename="%s"' % download

    stats = os.stat(filename)
    headers['Content-Length'] = clen = stats.st_size
    headers['Last-Modified'] = email.utils.formatdate(stats.st_mtime, usegmt=True)

    ims = env_get('HTTP_IF_MODIFIED_SINCE')
    if ims:
        ims = parse_date(ims.split(";")[0].strip())
    if ims is not None and ims >= int(stats.st_mtime):
        headers['Date'] = email.utils.formatdate(time.time(), usegmt=True)
        return HTTPResponse(status=304, **headers)

    body = '' if request.method == 'HEAD' else open(filename, 'rb')

    headers["Accept-Ranges"] = "bytes"
    range_header = env_get('HTTP_RANGE')
    if range_header:
        first_range = get_first_range(range_header, clen)
        if not first_range:
            return HTTPError(416, "Requested Range Not Satisfiable")
        offset, end = first_range
        headers["Content-Range"] = f"bytes {offset}-{end-1}/{clen}"
        headers["Content-Length"] = str(end - offset)
        if body:
            body = _file_iter_range(body, offset, end - offset)
        return HTTPResponse(body, status=206, **headers)
    return HTTPResponse(body, **headers)



from collections.abc import MutableMapping as DictMixin
import types
import threading
import base64
import pickle
import hmac
import hashlib
import email.utils
import time
from . import errors


def parse_date(ims):
    """ Parse rfc1123, rfc850 and asctime timestamps and return UTC epoch. """
    try:
        ts = email.utils.parsedate_tz(ims)
        return time.mktime(ts[:8] + (0,)) - (ts[9] or 0) - time.timezone
    except (TypeError, ValueError, IndexError, OverflowError):
        return None


# ---------------- [ Some helpers for string/byte handling ] ---------
def tob(s, enc='utf8'):
    encode = getattr(s, 'encode', None)
    if encode:
        return encode(enc)
    return bytes(s)


def touni(s, enc='utf8', err='strict'):
    decode = getattr(s, 'decode', None)
    if decode:  # bytes
        return decode(enc, err)
    return str(s)


# ------------------[ cookie ] -------------------
def cookie_encode(data, key):
    ''' Encode and sign a pickle-able object. Return a (byte) string '''
    msg = base64.b64encode(pickle.dumps(data, -1))
    sig = base64.b64encode(hmac.new(tob(key), msg, digestmod=hashlib.md5).digest())
    return tob('!') + sig + tob('?') + msg


def cookie_decode(data, key):
    ''' Verify and decode an encoded string. Return an object or None.'''

    def cookie_is_encoded(data):
        ''' Return True if the argument looks like a encoded cookie.'''
        return bool(data.startswith(tob('!')) and tob('?') in data)

    def _lscmp(a, b):
        ''' Compares two strings in a cryptographically safe way:
            Runtime is not affected by length of common prefix. '''
        return not sum(0 if x == y else 1 for x, y in zip(a, b)) and len(a) == len(b)

    data = tob(data)
    if cookie_is_encoded(data):
        sig, msg = data.split(tob('?'), 1)
        if _lscmp(sig[1:], base64.b64encode(hmac.new(tob(key), msg, digestmod=hashlib.md5).digest())):
            return pickle.loads(base64.b64decode(msg))
    return None


# ------------------[ html escape `&<>'"`] -------------------
def html_escape(string):
    ''' Escape HTML special characters ``&<>`` and quotes ``'"``. '''
    return string.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;')\
                 .replace('"', '&quot;').replace("'", '&#039;')


# ------------------[ thread safe props] -------------------
def ts_props(*props, store_name=None):
    def wrapper(cls):
        local_store = None
        cls_init = cls.__init__

        def init_wrapper(self, *a, **kw):
            nonlocal local_store
            local_store = getattr(self, store_name, None)
            if local_store is None:
                local_store = threading.local()
                setattr(self, store_name, local_store)
            [setattr(local_store, k, None) for k in props]
            cls_init(self, *a, **kw)

        def make_prop(k):
            def fget(s):
                return getattr(local_store, k)

            def fset(s, v):
                return setattr(local_store, k, v)

            def fdel(s):
                return delattr(local_store, k)
            doc = 'Local property: %s' % k
            return property(fget, fset, fdel, doc)

        cls.__init__ = init_wrapper
        [setattr(cls, p, make_prop(p)) for p in props]
        return cls
    return wrapper


# ------------------[ exposes prop callable attrubutes at instance level] -------------------
def proxy(prop, attrs, cls = None):
    def injector(cls):
        for attr in attrs:
            setattr(
                cls, attr,
                (lambda _attr = attr: lambda s, *a, **kw: getattr(getattr(s, prop), _attr)(*a, **kw))()
            )
        return cls
    return injector if not cls else injector(cls)


# ------------------[ helper classes] -------------------

class cached_property(object):
    ''' A property that is only computed once per instance and then replaces
        itself with an ordinary attribute. Deleting the attribute resets the
        property. '''

    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self
        try:
            value = self.func(obj)
        except AttributeError as err:
            raise errors.PropertyGetterError(
                'AttributeError in cached_property getter of '
                f'`{self.func.__name__}`: {str(err)}'
            )
        setattr(obj, self.func.__name__, value)
        return value


class NameSpace(types.SimpleNamespace):
    ''' fast Name-Space-Dict
    nsd.some       # - fast
    nsd['some']    # - 20...30% slower
    '''
    __getitem__ = types.SimpleNamespace.__getattribute__
    __setitem__ = types.SimpleNamespace.__setattr__
    get = lambda s, k, d=None: s.__dict__.get(k, d)
    keys = lambda s: s.__dict__.keys()
    values = lambda s: s.__dict__.values()
    items = lambda s: s.__dict__.items()
    setdefault = lambda s, k, d: s.__dict__.setdefault(k, d)
    update = lambda s, d: s.__dict__.update(d)

    def __init__(self, **kw):
        super().__init__(**kw)


class _MetaSimpleConfig(type):

    def __init__(cls, name, bases, dct):
        keys = cls.__get_keys__(bases)
        if keys:
            for k in dct.keys():
                if k.startswith('_'):
                    continue
                if k not in keys:
                    raise KeyError(f'Unexpected key: {k}')

    @staticmethod
    def __get_keys__(bases):
        ret_keys = None
        for bcls in bases:
            keys = getattr(bcls, '__keys__', None)
            if not keys:
                continue
            if not ret_keys:
                ret_keys = keys
            elif keys != ret_keys:
                raise TypeError('Multiple keys holders detected')
        return ret_keys


class SimpleConfig(metaclass=_MetaSimpleConfig):

    @classmethod
    def keys_holder(cls, holder_cls):
        assert cls.__base__ is object
        keys_holder_cls = getattr(holder_cls, '__keys_holder__', None)
        if keys_holder_cls:
            raise RuntimeError(f'Keys holder is already registered: {keys_holder_cls}')

        keys = set(holder_cls.keys())
        for k in keys:
            if hasattr(cls, k):
                raise KeyError(f'Bad key `{k}`, reserved keys/attrs are {cls.keys()}')
        holder_cls.__keys__ = keys
        holder_cls.__keys_holder__ = holder_cls
        return holder_cls

    def __new__(cls, src_config=None, **kw):
        return cls.get_from(src_config, **kw)

    @classmethod
    def keys(cls):
        keys = getattr(cls, '__keys__', None)
        if keys:
            return keys.copy()
        return (k for k in cls.__dict__ if not k.startswith('__'))

    @classmethod
    def items(cls):
        return ((k, getattr(cls, k)) for k in cls.keys())

    @classmethod
    def get_from(cls, src_config=None, **kw):
        if src_config is None:
            src_config = {}
        return NameSpace(**{
            key: src_config.get(key, kw.get(key, default))
            for key, default in cls.items()
        })

    @classmethod
    def get(cls, k, default=None):
        return getattr(cls, k) if k in cls.keys() else default


# ------------------[ headers parsing] -------------------
def _hval(value):
    if not (value is None or isinstance(value, (str, int, float, bool))):
        raise TypeError(f"Header value must be type of (str, int, float, bool, None), got: {type(value)}")
    value = str(value)
    if '\n' in value or '\r' in value or '\0' in value:
        raise ValueError("Header value must not contain control characters: %r" % value)
    return value


@proxy('dict', 'keys pop popitem values items get'.split())
class HeaderDict(DictMixin):
    __slots__ = ('_ts',)

    dict = property(
        (lambda s: s._ts.dict),
        (lambda s, v: setattr(s._ts, 'dict', v)),
    )

    def __init__(self, *a, **kw):
        self._ts = threading.local()
        self._ts.dict = dict(*a, **kw)

    def __len__(self):
        return len(self._ts.dict)

    def __iter__(self):
        return self._ts.dict.__iter__()

    def __contains__(self, key):
        return key in self._ts.dict

    def __delitem__(self, key):
        del self._ts.dict[key]

    def __getitem__(self, key):
        return self._ts.dict[key]

    def __setitem__(self, key, value):
        self._ts.dict[key] = _hval(value)

    def copy(self):
        ret = self.__class__()
        ret.dict = {k: (v[:] if isinstance(v, list) else v) for k, v in self.items()}
        return ret

    def setdefault(self, key, value):
        return self._ts.dict.setdefault(key, _hval(value) if not isinstance(value, list) else value)

    def append(self, key, value):
        d = self._ts.dict
        value = _hval(value)
        v = d.get(key)
        if v is None:
            d[key] = value
        elif isinstance(v, list):
            v.append(value)
        else:
            d[key] = [v, value]

    def clear(self, *names):
        if names:
            for n in names:
                if n in self:
                    del self[n]
        else:
            self._ts.dict.clear()

    def update(self, d):
        self._ts.dict.update(d)

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self.dict}>'


class HeaderProperty:

    __slots__ = ('name', 'default', 'reader', 'writer', '__doc__')

    def __init__(self, name, reader=None, writer=None, default=''):
        self.name, self.default = name, default
        self.reader, self.writer = reader, writer
        self.__doc__ = 'Current value of the %r header.' % name.title()

    def __get__(self, obj, cls):
        if obj is None:
            return self
        value = obj.headers.get(self.name, self.default)
        return self.reader(value) if self.reader else value

    def __set__(self, obj, value):
        obj.headers[self.name] = self.writer(value) if self.writer else value

    def __delete__(self, obj):
        del obj.headers[self.name]


class WSGIFileWrapper(object):

    def __init__(self, fp, buffer_size=1024 * 64):
        self.fp, self.buffer_size = fp, buffer_size
        for attr in ('fileno', 'close', 'read', 'readlines', 'tell', 'seek'):
            a = getattr(fp, attr, None)
            if a is not None:
                setattr(self, attr, a)

    def __iter__(self):
        buff, read = self.buffer_size, self.read
        part = read(buff)
        while part:
            yield part
            part = read(buff)



import os
from collections.abc import MutableMapping as DictMixin
import re
from unicodedata import normalize
from urllib.parse import unquote as _urlunquote
from functools import partial

from ..common_helpers import touni, HeaderDict, HeaderProperty, cached_property
from .. import errors


urlunquote = partial(_urlunquote, encoding='latin1')


# ------------------[ cachable property with a custom storage ] -------------------
def cache_in(attr, key=None, read_only=False):
    # attr = 'environ[ PATH_INFO ]'
    re_attr_key = re.compile(r'^(.+?)\[\s*([^\[\]]+?)\s*\]$')
    if not key:
        attr_key = re_attr_key.match(attr)
        if attr_key:
            attr, key = attr_key.groups()

    def wrapper(getter):

        if not key:
            def fget(self):
                try:
                    return getattr(self, attr)
                except AttributeError:
                    try:
                        v = getter(self)
                    except AttributeError as err:
                        raise errors.PropertyGetterError(
                            f'AttributeError in getter of cache_in property `{getter.__name__}`, '
                            f'storage={attr}: {str(err)}'
                        )
                    setattr(self, attr, v)
                    return getattr(self, attr)

            def fset(self, value):
                if read_only:
                    raise AttributeError("Read-Only property.")
                setattr(self, attr, value)

            def fdel(self):
                if read_only:
                    raise AttributeError("Read-Only property.")
                delattr(self, attr)
        else:
            def fget(self):
                storage = getattr(self, attr)
                if key not in storage:
                    try:
                        storage[key] = getter(self)
                    except AttributeError as err:
                        raise errors.PropertyGetterError(
                            f'AttributeError in getter of cache_in property `{getter.__name__}`, '
                            f'storage={attr}[{key}]: {str(err)}'
                        )
                return storage[key]

            def fset(self, value):
                if read_only:
                    raise AttributeError("Read-Only property.")
                getattr(self, attr)[key] = value

            def fdel(self):
                if read_only:
                    raise AttributeError("Read-Only property.")
                del getattr(self, attr)[key]

        return property(fget, fset, fdel, 'cache_in')
    return wrapper


def parse_qsl(qs, *, append: callable = None, setitem: callable = None):
    container = None
    if setitem:
        _seen = dict()
        _lists = dict()

        def add(k, v):
            vlist = _lists.get(k)
            if vlist:
                vlist.append(v)
            elif k in _seen:
                tmp = _lists[k] = [_seen[k], v]
                setitem(k, tmp)
            else:
                setitem(k, _seen.setdefault(k, v))
    elif append:
        add = lambda k, v: append((k, v))
    else:
        container = []
        _append = container.append
        add = lambda k, v: _append((k, v))

    L = len(qs)
    i = 0
    while i < L:
        key = None
        idx = 0; c = None
        for idx, c in enumerate(qs[i:]):
            if c == '=' or c == '&':
                break
        else:
            idx += 1
        j = i + idx
        key = qs[i:j]
        i = j + 1  # skip '=' or '&'
        if not key:
            continue
        key = urlunquote(key.replace('+', ' '))
        if c == '&':
            value = ''
        else:
            idx = 0; c = None
            for idx, c in enumerate(qs[i:]):
                if c == '&':
                    break
            else:
                idx += 1
            j = i + idx
            value = urlunquote(qs[i:j].replace('+', ' '))
            i = j + 1  # skip '&'
        add(key, value)
    return container


class FormsDict(dict):
    ''' This :class:`dict` subclass is used to store request form data.
        Additionally to the normal dict-like item access methods (which return
        unmodified data as native strings), this container also supports
        attribute-like access to its values. Attributes are automatically de-
        or recoded to match :attr:`input_encoding` (default: 'utf8'). Missing
        attributes default to an empty string. '''

    #: Encoding used for attribute values.
    input_encoding = 'utf8'
    #: If true (default), unicode strings are first encoded with `latin1`
    #: and then decoded to match :attr:`input_encoding`.
    recode_unicode = True

    def _fix(self, s, encoding=None):
        if isinstance(s, str) and self.recode_unicode:  # Python 3 WSGI
            return s.encode('latin1').decode(encoding or self.input_encoding)
        else:
            return s

    def decode(self, encoding=None):
        ''' Returns a copy with all keys and values de- or recoded to match
            :attr:`input_encoding`. Some libraries (e.g. WTForms) want a
            unicode dictionary. '''
        copy = FormsDict()
        enc = copy.input_encoding = encoding or self.input_encoding
        copy.recode_unicode = False
        for key, value in self.items():
            copy[self._fix(key, enc)] = self._fix(value, enc)
        return copy

    def getunicode(self, name, default=None, encoding=None):
        ''' Return the value as a unicode string, or the default. '''
        try:
            return self._fix(self[name], encoding)
        except (UnicodeError, KeyError):
            return default

    def __getattr__(self, name, default=str()):
        # Without this guard, pickle generates a cryptic TypeError:
        if name.startswith('__') and name.endswith('__'):
            return super().__getattr__(name)
        return self.getunicode(name, default=default)


class WSGIHeaderDict(DictMixin):
    ''' This dict-like class wraps a WSGI environ dict and provides convenient
        access to HTTP_* fields. Keys and values are native strings
        (2.x bytes or 3.x unicode) and keys are case-insensitive. If the WSGI
        environment contains non-native string values, these are de- or encoded
        using a lossless 'latin1' character set.
    '''
    #: List of keys that do not have a ``HTTP_`` prefix.
    cgikeys = {'CONTENT_TYPE', 'CONTENT_LENGTH'}
    __slots__ = ('environ',)

    def __init__(self, environ):
        self.environ = environ

    def _ekey(self, key):
        ''' Translate header field name to CGI/WSGI environ key. '''
        key = key.replace('-', '_').upper()
        if key in self.cgikeys:
            return key
        return 'HTTP_' + key

    def raw(self, key, default=None):
        ''' Return the header value as is (may be bytes or unicode). '''
        return self.environ.get(self._ekey(key), default)

    def __getitem__(self, key):
        return touni(self.environ[self._ekey(key)], 'latin1')

    def __setitem__(self, key, value):
        raise TypeError("%s is read-only." % self.__class__)

    def __delitem__(self, key):
        raise TypeError("%s is read-only." % self.__class__)

    def __iter__(self):
        for key in self.environ:
            if key.startswith('HTTP_'):
                yield key[5:].replace('_', '-').title()
            elif key in self.cgikeys:
                yield key.replace('_', '-').title()

    def keys(self): return [x for x in self]
    def __len__(self): return len(self.keys())
    def __contains__(self, key): return self._ekey(key) in self.environ


class FileUpload:

    __slots__ = ('file', 'name', 'raw_filename', 'headers', '__dict__')

    content_type = HeaderProperty('Content-Type')
    content_length = HeaderProperty('Content-Length', reader=int, default=-1)

    def __init__(self, fileobj, name, filename, headers=None):
        ''' Wrapper for file uploads. '''
        #: Open file(-like) object (BytesIO buffer or temporary file)
        self.file = fileobj
        #: Name of the upload form field
        self.name = name
        #: Raw filename as sent by the client (may contain unsafe characters)
        self.raw_filename = filename
        #: A :class:`HeaderDict` with additional headers (e.g. content-type)
        self.headers = HeaderDict(headers) if headers else HeaderDict()

    def get_header(self, name, default=None):
        """ Return the value of a header within the multipart part. """
        return self.headers.get(name, default)

    @cached_property
    def filename(self):
        ''' Name of the file on the client file system, but normalized to ensure
            file system compatibility. An empty filename is returned as 'empty'.

            Only ASCII letters, digits, dashes, underscores and dots are
            allowed in the final filename. Accents are removed, if possible.
            Whitespace is replaced by a single dash. Leading or tailing dots
            or dashes are removed. The filename is limited to 255 characters.
        '''
        fname = self.raw_filename
        if not isinstance(fname, str):
            fname = fname.decode('utf8', 'ignore')
        fname = normalize('NFKD', fname).encode('ASCII', 'ignore').decode('ASCII')
        fname = os.path.basename(fname.replace('\\', os.path.sep))
        fname = re.sub(r'[^a-zA-Z0-9-_.\s]', '', fname).strip()
        fname = re.sub(r'[-\s]+', '-', fname).strip('.-')
        return fname[:255] or 'empty'

    def _copy_file(self, fp, chunk_size=2**16):
        read, write, offset = self.file.read, fp.write, self.file.tell()
        while True:
            buf = read(chunk_size)
            if not buf:
                break
            write(buf)
        self.file.seek(offset)

    def save(self, destination, overwrite=False, chunk_size=2**16):
        ''' Save file to disk or copy its content to an open file(-like) object.
            If *destination* is a directory, :attr:`filename` is added to the
            path. Existing files are not overwritten by default (IOError).

            :param destination: File path, directory or file(-like) object.
            :param overwrite: If True, replace existing files. (default: False)
            :param chunk_size: Bytes to read at a time. (default: 64kb)
        '''
        if isinstance(destination, str):  # Except file-likes here
            if os.path.isdir(destination):
                destination = os.path.join(destination, self.filename)
            if not overwrite and os.path.exists(destination):
                raise IOError('File exists.')
            with open(destination, 'wb') as fp:
                self._copy_file(fp, chunk_size)
        else:
            self._copy_file(destination, chunk_size)



import re

from .sym_stream import SymStream
from .errors import RouteSyntaxError


class Parser:
    param_delimiters = '{<'
    param_delimiters_map = {
        '<': '>',
        '{': '}'
    }

    def __init__(self):
        self._stream = None
        self._rule = None

    def iter_parse(self, rule):
        self._rule = rule
        self._stream = SymStream(rule)
        yield from self._iter_parse()
        self._rule = None
        self._stream = None

    def _iter_parse(self):
        '''
        valid examples:
            `/abc/:some/foo` - simple param (param name matches all from `:` to `/`)
            `/abc/<some>other/foo` - simple param (param name matches all from `:` to `/`)
            `/abc/<param.path()>baz/foo` - matches `/abc/a/b/cbaz/foo`, param is `a/b/c`
            `/abc/<param.rex((foo)|(bar)|(baz))[1]>/foo` - matches `/abc/foobaz/foo`, param is `foo`
            `/abc/<param.rex((foo)|(bar)|(baz))[2]>baz/foo`   - matches `/abc/barbaz/foo`, param is  `bar`
            `/abc/<rex((foo)|(bar)|(baz))[3]>baz/foo`   - matches `/abc/bazbaz/foo`
        '''

        S = self._stream

        param_tokens = self.param_delimiters + ':'
        while S.current:
            part, param, filter, filter_args, filter_selector = (None,) * 5
            # param logic
            if S.current in param_tokens:
                param, filter, filter_args, filter_selector = self._parse_param()
                # path needs special processing
                if filter == 'path':
                    tail = S.rest()
                    tmp = re.match(fr'[^{param_tokens}]+', tail)
                    token_pos = tmp.end() if tmp else len(tail)
                    filter_args = tail[: token_pos]
            # static logic
            else:
                part = S.eat(fr'[^{param_tokens}]+')

            yield part, param, filter, filter_args, filter_selector

    def _parse_param(self):
        S = self._stream
        rule = self._rule

        is_bottle_filter = False
        param, filter, filter_args, filter_selector = (None,) * 4
        py_name = r'[a-zA-Z_]\w*'
        first = S.current

        if first == ':':
            S.next()  # eat ':'
            if S.current is not None:
                param = S.expect(fr'({py_name})?((?=/)|$)', group = 1) or None
        else:
            assert first in self.param_delimiters
            dclose = self.param_delimiters_map[first]
            S.next()  # eat open-delimiter
            if S.current == ':':  # bottle styled filter
                S.next()
                is_bottle_filter = True

            name = S.expect(fr'{py_name}')
            if is_bottle_filter:
                filter = name

            if S.current == dclose:
                if not filter:
                    param = name
            elif S.current == '.':
                S.next()  # eat '.' or ':'
                param = name
                filter = S.expect(fr'{py_name}')
            elif S.current == ':':
                if not filter:
                    S.next()  # eat '.' or ':'
                    param = name
                    filter = S.expect(fr'{py_name}')
            elif S.current == '(':
                if not filter:
                    filter = name
            else:
                raise RouteSyntaxError(
                    f'Unexpected syntax in `...{S.rest()}`\n\tin {rule}'
                )

            if filter:
                if S.current not in f':({dclose}':
                    raise RouteSyntaxError(
                        f'Expected one of `(,:,>` after filter instead of `...{S.rest()}`\n\tin {rule}'
                    )
                if S.current != dclose:
                    if S.current == '(':
                        filter_args = S.expect_parenthesized()[1:-1]
                        if S.current == '[':
                            filter_selector = S.expect(r'\[(.+?)\]', 1)
                            if not filter_selector:
                                raise RouteSyntaxError(
                                    f'filter selector value was expected instead of `...{S.rest()}`\n\tin {rule}'
                                )
                    else:  # bottle style
                        S.next()
                        filter_args = S.eat(fr'[^{dclose}]+')
            S.expect(dclose)
        return param, filter, filter_args, filter_selector



from ..errors import OmbottException


class RouteError(OmbottException):
    """ This is a base class for all routing related exceptions """
    __slots__ = ('extra',)

    def __init__(self, *args, **kw):
        super().__init__(*args)
        self.extra = kw


class RouteSyntaxError(RouteError):
    """ The route parser found something not supported by this router. """


class RouteBuildError(RouteError):
    """ The route could not be built. """


class RouteMethodError(RouteError):
    """ Method not allowed """


#!/usr/bin/env python3

import os
from enum import IntEnum
common_pref = os.path.commonprefix

__author__ = "Valery Kucherov <valq7711@gmail.com>"
__copyright__ = "Copyright (C) 2021 Valery Kucherov"
__license__ = "MIT"
__version__ = "0.0.3"


KEY = 0
IDX = 1
WEIGHT = 2  # doesn`t use at the moment
PARAMS = 3
FILTER = 4
IS_EXCLUSIVE = 5
HOOKS = 6
DATA = 7
OFFSET = 8


# mismatch types
class MismatchType(IntEnum):
    WHOLE = 1
    PARTIAL = 2
    TOKEN = 3
    FILTER = 4
    EXCLUSIVITY = 5


class RadiDictError(Exception):
    pass


class RadiDictKeyError(RadiDictError):
    def __init__(self, msg, **kw):
        super().__init__(msg)
        self.__err_args__ = kw

    def __getattr__(self, a):
        return self.__err_args__[a]


class RadiDict:
    def __init__(
        self, *,
        path_sep = '/',
        param_token = '\r',
        is_exclusive = False
    ):
        self.path_sep = path_sep
        self.param_token = param_token
        self.is_exclusive = is_exclusive
        self.root = self._make_node(key = '/')

    def _token_pos(self, s):
        ret = s.find(self.param_token)
        return ret if ret >= 0 else len(s)

    def _make_node(self, key, *,
                   idx = None, data = None, weight = 0,
                   children = None, params = None, filter = None,
                   is_exclusive = False, hooks = None):
        return [
            key,
            idx,
            weight,
            params or [],
            filter,
            is_exclusive,
            hooks,
            data,
            *(children or [])
        ]

    def _mount(self, pnode, child):

        keyerror = RadiDictKeyError
        ckey = child[KEY]

        if pnode[IDX]:
            has_token = pnode[IDX][-1] == self.param_token and pnode[IDX][-1]
            # if pnode has exclusive token child
            # it will be only one child at pnode[OFFSET]
            has_exclusive = has_token and pnode[OFFSET][IS_EXCLUSIVE]
            if has_exclusive:
                raise keyerror('can`t add `{ckey}` since exlusive token `{ptoken}` is already here')
            elif has_token and ckey == has_token:
                raise keyerror('can`t add `{ckey}` since token `{ptoken}` is already here')
            # maybe child is exclusive token
            elif child[IS_EXCLUSIVE]:
                raise keyerror('can`t add `{ckey}` since it is exclusive token but some keys are already here')
        # no children
        else:
            pnode[IDX] = ''

        if ckey == self.param_token:
            pnode.append(child)
            pnode[IDX] += ckey
        else:
            pnode.insert(OFFSET, child)
            pnode[IDX] = ckey[0] + pnode[IDX]
        pnode[WEIGHT] += 1
        return child

    def _split(self, pnode, split_idx: int = None, *, by_key = None):
        keyerror = RadiDictKeyError
        key = pnode[KEY]
        if by_key:
            split_idx = len(common_pref([key, by_key]))
            if not by_key or not key[split_idx:]:
                raise keyerror('something went wrong')
        new_key = key[:split_idx]
        key_rest = key[split_idx:]
        node = pnode[:]
        pnode[:] = self._make_node(
            key = new_key,
            idx = key_rest[0],
            weight = node[WEIGHT] + 1,
            children = [node]
        )
        node[KEY] = key_rest
        return pnode if not by_key else (pnode, split_idx)

    def _try_merge(self, pnode):
        if (
            pnode is self.root
            or pnode[DATA] or pnode[HOOKS]
            or not pnode[IDX] or len(pnode[IDX]) != 1
            or pnode[KEY] == self.param_token
            or pnode[IDX] == self.param_token
        ):  # i.e. child is token
            return

        child = pnode[OFFSET]
        new_key = pnode[KEY] + child[KEY]
        pnode[:] = child
        pnode[KEY] = new_key

    def _make_route(self, pnode, route_pattern, data, hooks, param_exclusions, param_filters, param_names):
        '''
        if route_pattern is sliced  param_exclusions, param_filters should be also sliced
        '''

        route = route_pattern
        TOKEN = self.param_token

        i = 0
        param_idx = 0
        while i < len(route):
            args = {}  # new node args
            route_rest = route[i:]
            key_len = route_rest.find(TOKEN)
            if key_len == 0:
                route_key = TOKEN
                key_len = 1
                args['is_exclusive'] = param_exclusions[param_idx]
                args['filter'] = param_filters[param_idx]
                # param_idx += 1
            elif key_len > 0:
                route_key = route_rest[: key_len]
            # no tokens
            else:
                route_key = route_rest
                key_len = len(route_key)

            args['key'] = route_key
            try:
                pnode = self._mount(pnode, self._make_node(**args))
                if route_key == TOKEN:
                    param_idx += 1
            except RadiDictKeyError as e:
                e.__err_args__['param_idx'] = param_idx
                raise

            i += key_len

        pnode[DATA] = data
        pnode[HOOKS] = hooks or None
        pnode[PARAMS] = param_names

    @staticmethod
    def params_unpack(params: dict):
        exclusions = []
        filters = []
        keys = []
        if params:
            for _e, _f in params.values():
                exclusions.append(_e)
                filters.append(_f)
            keys = list(params.keys())
        return exclusions, filters, keys

    def _set(self, pnode, route_pattern,
             *, data = None, hooks = None, params: dict = None, overwrite = True):
        '''
        params = {name:str :  [is_exclusive:boolean, filter:callable]}
        '''

        def error(msg):
            args = dict(
                matched = self._render_route(route_pattern[:ptr], prm_keys),
                route = self._render_route(route_pattern, prm_keys),
                raw_route = route_pattern,
                params = params,
                data = data,
                node = node
            )
            msg = msg.format(**args) + '\n ' + '\n  '.join([
                'matched:   {matched}',
                'route:     {route}'
            ]).format(**args)
            raise RadiDictKeyError(msg, **args)

        route = route_pattern
        prm_exclusions, prm_filters, prm_keys = self.params_unpack(params)
        node, mismatch, ptr, prm_idx, _ = self._match(route, prm_exclusions, prm_filters, pnode)
        if mismatch == MismatchType.EXCLUSIVITY:
            error('tokens exclusivity mismatch')
        elif mismatch == MismatchType.FILTER:
            error('tokens filter mismatch')
        elif mismatch == MismatchType.PARTIAL:
            ckey = route[ptr:]
            ckey = ckey[: self._token_pos(ckey)]
            node, split_idx = self._split(node, by_key=ckey)
            ptr += split_idx
            mismatch = False if ptr >= len(route) else MismatchType.WHOLE

        if not mismatch:
            for item_idx, item in [(DATA, data), (HOOKS, hooks)]:
                if item is None:
                    continue
                if node[item_idx] and not overwrite:
                    error('handler is already registered here')
                else:
                    node[item_idx] = item
            if data is not None:  # prevent to overwrite param keys by hooks
                node[PARAMS] = prm_keys
            return
        else:
            try:
                self._make_route(
                    node, route[ptr:], data, hooks,
                    prm_exclusions[prm_idx:], prm_filters[prm_idx:], prm_keys
                )
            except RadiDictKeyError as e:
                msg = e.args[0]
                msg = msg.format(
                    ckey = prm_keys[prm_idx + e.param_idx],
                    ptoken = '<param>'
                )
                error(f'Key error: {msg}')

    def _match(self, route_pattern, param_exclusions: list = None, param_filters: list = None, pnode = None):
        pnode = pnode if pnode is not None else self.root
        route = route_pattern
        TOKEN: str = self.param_token
        L: int = len(route)
        i = 0
        param_idx = 0
        mismatch = None
        key_end = 0
        stack = [pnode]
        is_exact = False
        while i < L:
            pnode_idx = pnode[IDX]
            c0 = route[i]
            kidx = 0
            for ic, c in enumerate(pnode_idx or []):
                if c == c0:
                    kidx = ic; break  # found!
            else:
                break  # not found - break while

            pnode = pnode[kidx + OFFSET]
            key = pnode[KEY]
            key_end = i + len(key)
            stack.append(pnode)
            if key == route[i: key_end]:
                if key == TOKEN:
                    if (
                        param_exclusions
                        and pnode[IS_EXCLUSIVE] != param_exclusions[param_idx]
                    ):
                        mismatch = MismatchType.EXCLUSIVITY
                        break  # while: fail
                    elif (
                        param_filters
                        and pnode[FILTER] != param_filters[param_idx]
                    ):
                        mismatch = MismatchType.FILTER
                        break  # while: fail

                    param_idx += 1
                i = key_end
            else:
                mismatch = MismatchType.PARTIAL
                break  # while: fail
        else:  # success
            is_exact = True

        if not is_exact and not mismatch:
            mismatch = MismatchType.WHOLE
        return pnode, mismatch, i, param_idx, stack

    # ------------------- [interface methods] --------------------------
    def add(self, route_pattern, data, params = None, *, overwrite = False):
        if isinstance(params, (list, tuple)):
            params = dict.fromkeys(params, [self.is_exclusive, None])
        return self._set(
            self.root, route_pattern,
            data = data, params = params, overwrite = overwrite
        )

    def add_hooks(self, route_pattern, hooks, params = None, *, overwrite = False):
        if isinstance(params, (list, tuple)):
            params = dict.fromkeys(params, [self.is_exclusive, None])
        return self._set(
            self.root, route_pattern,
            hooks = hooks, params = params, overwrite = overwrite
        )

    def remove(self, route_pattern, hooks_only=False):

        is_wildcard = False
        if route_pattern and route_pattern[-1] == '*':
            route_pattern = route_pattern[:-1]
            is_wildcard = True
            if hooks_only:
                raise RadiDictError('Can`t remove hooks by wildcard')

        node, mismatch, ptr, prm_idx, stack = (
            self._match(route_pattern, pnode = self.root)
        )

        if is_wildcard and (
            not mismatch
            or mismatch == MismatchType.PARTIAL
            and node[KEY].startswith(route_pattern[ptr:])
        ):
            node[IDX] = None
            del node[OFFSET:]
        elif mismatch:
            return

        if hooks_only and node[DATA] is not None:
            node[HOOKS] = None
            return

        stack.reverse()
        assert node is stack[0]
        node[DATA] = None
        node[PARAMS] = []
        key0_to_del = None
        for node in stack:
            if key0_to_del:
                kidx = node[IDX].find(key0_to_del)
                if kidx < 0:
                    raise RadiDictError(
                        'critical error: router seems to be corrupted'
                    )
                node[IDX] = node[IDX].replace(key0_to_del, '')
                del node[OFFSET + kidx]
                self._try_merge(node)
            if not (node[DATA] or node[IDX]):
                key0_to_del = node[KEY][0]
            else:
                break

    def get(self, route, allow_partial=False):
        # local constants
        TOKEN: str = self.param_token
        PATH_SEP: str = self.path_sep
        L: int = len(route)

        pnode = self.root
        params = []
        hooks = pnode[HOOKS]
        hooks = [[0, hooks]] if hooks else []
        look_back = []
        do_look_back = False
        selector = None
        i = 0
        while True:
            while i < L:
                idx = pnode[IDX]
                if not idx:
                    # there are no children
                    break  # while

                kidx = None
                if do_look_back:
                    do_look_back = False
                    c = idx[-1]
                else:
                    c0 = route[i]
                    for ic, c in enumerate(idx):
                        if c == c0:
                            kidx = ic; break  # found!

                if kidx is None:  # not found
                    # maybe token or look_back
                    if c == TOKEN:
                        token_node = pnode[-1]
                        filter = token_node[FILTER]
                        if filter:
                            param_value, j, selector = filter(route[i:])
                            j += i
                        else:
                            j = i
                            while j < L:
                                if route[j] == PATH_SEP:
                                    break
                                j += 1
                            param_value = route[i:j]

                        if param_value is not None:
                            params.append(param_value)
                            i = j
                            pnode = token_node
                            h = pnode[HOOKS]
                            if h:
                                hooks.append([i, h])
                            if selector is not None:
                                look_back.append(
                                    [route, pnode, i, L, params[:], hooks[:], False]
                                )
                                route = str(selector) + route[i:]
                                i = 0
                                L = len(route)
                                selector = None
                            continue
                        else:
                            break  # while
                    else:
                        break

                # regular logic
                if idx[-1] == TOKEN:
                    look_back.append(
                        [route, pnode, i, L, params[:], hooks[:], True]
                    )
                pnode = pnode[OFFSET + kidx]
                key = pnode[KEY]
                key_end = i + len(key)
                if key == route[i: key_end]:
                    i = key_end
                    h = pnode[HOOKS]
                    if h:
                        hooks.append([i, h])
                else:
                    break  # while

            else:  # while-postprocessing
                if pnode[DATA]:
                    return (
                        pnode[DATA],
                        dict(
                            param_keys = pnode[PARAMS],
                            param_values = params,
                            hooks = hooks
                        )
                    )
                # if there is no data
                # this is partial match => try look_back

            if look_back:
                route, pnode, i, L, params, hooks, look_type = look_back.pop()
                do_look_back = look_type
            else:
                return None if not allow_partial else (
                    None,
                    dict(
                        param_values = params,
                        hooks = hooks,
                        partial = route[:i]
                    )
                )

    # ---------------- [helpers] --------------------------
    def _render_route(self, route, params):
        if not params:
            return route
        for p in params:
            route = route.replace(self.param_token, f':{p or "noname"}', 1)
        return route

    def _routes_iter(self, pnode = None, startswith = None, yield_hooks = False):
        pnode = pnode or self.root
        prefix_stack = []
        if startswith:
            node, mismatch, ptr, _, prefix_stack = self._match(startswith, pnode = pnode)
            if (
                not mismatch
                or mismatch == MismatchType.PARTIAL
                and node[KEY].startswith(startswith[ptr:])
            ):
                pnode = node
                del prefix_stack[-1]
            else:
                return []

        rec = [
            pnode,  # path
            len(pnode[IDX] or ''),
            0,  # cur child index
        ]
        path = [pnode]
        stack = [rec]
        while stack:
            pnode, L, i = stack[-1]
            if i < L:
                cnode = pnode[i + OFFSET]
                L = len(cnode[IDX] or '')
                stack[-1][2] = i + 1
                path.append(cnode)
                stack.append([
                    cnode,
                    L,
                    0
                ])
            else:
                if pnode[DATA] or yield_hooks and pnode[HOOKS]:
                    yield prefix_stack + path
                stack.pop()
                path.pop()



from pathlib import Path

html = Path(__file__).parent / 'error.html'

_html_lns = []

def render(err_resp, url, debug):
    ex = traceback = '-] Forbidden [-'
    if debug:
        traceback = err_resp.traceback
        ex = err_resp.exception
        try:
            ex = repr(ex)
        except:
            ex = f'<unprintable {type(ex)} object>'
    ctx = dict(
        e = err_resp,
        exception = ex,
        traceback = traceback,
        url = repr(url)
    )

    if not _html_lns:
        with html.open('r') as f:
            s = f.readlines()
        _html_lns[:] = [ln.strip() for ln in s]

    out = []
    skip_until = ''
    for ln in _html_lns:
        if skip_until:
            if ln.startswith(skip_until):
                skip_until = ''
        elif ln.startswith('<style'):
            skip_until = '</style'
        else:
            ln = ln.format(**ctx)
        out.append(ln)
    return ''.join(out)


def _test():
    class e:
        status = 'status'
        body = 'body'
        traceback = 'traceback'
        exception = 'exception'

    print(render(e, 'url', False))





from datetime import date as datedate, datetime, timedelta
import email.utils
import http.client as httplib
from http.cookies import SimpleCookie
import time
import calendar

from .common_helpers import (
    ts_props,
    touni,
    cookie_encode,
    parse_date,
    HeaderDict, HeaderProperty,
)
from .errors import OmbottException


def http_date(value):
    if isinstance(value, str):
        return value
    if isinstance(value, datetime):
        # aware datetime.datetime is converted to UTC time
        # naive datetime.datetime is treated as UTC time
        value = value.utctimetuple()
    elif isinstance(value, datedate):
        # datetime.date is naive, and is treated as UTC time
        value = value.timetuple()
    if not isinstance(value, (int, float)):
        # convert struct_time in UTC to UNIX timestamp
        value = calendar.timegm(value)
    return email.utils.formatdate(value, usegmt=True)


_response_slots = (
    '_status_line', '_status_code', 'headers', '_headers',
    '_cookies', 'body', '_ts_props'
)


class BaseResponse:
    """ Storage class for a response body as well as headers and cookies.

        This class does support dict-like case-insensitive item-access to
        headers, but is NOT a dict. Most notably, iterating over a response
        yields parts of the body and not the headers.

        :param body: The response body as one of the supported types.
        :param status: Either an HTTP status code (e.g. 200) or a status line
                       including the reason phrase (e.g. '200 OK').
        :param headers: A dictionary or a list of name-value pairs.

        Additional keyword arguments are added to the list of headers.
        Underscores in the header name are replaced with dashes.
    """

    headers: HeaderDict = None

    __slots__ = ()

    default_status = 200
    default_content_type = 'text/html; charset=UTF-8'

    # Header blacklist for specific response codes
    # (rfc2616 section 10.2.3 and 10.3.5)
    bad_headers = {
        204: {'Content-Type'},
        304: {'Allow', 'Content-Encoding', 'Content-Language',
              'Content-Length', 'Content-Range', 'Content-Type',
              'Content-Md5', 'Last-Modified'}
    }

    def __new__(cls, *a, **kw):
        self = super().__new__(cls, *a, **kw)
        self.headers = HeaderDict()
        return self

    def __init__(self, body='', status=None, headers=None, **more_headers):
        self._status_line = None
        self._status_code = None
        self._cookies = None
        self._headers = {}
        self.headers.dict = self._headers
        self.body = body
        self.status = status or self.default_status
        if headers:
            if isinstance(headers, dict):
                headers = headers.items()
            for name, value in headers:
                self.headers.append(name, value)

        if more_headers:
            for name, value in more_headers.items():
                self.headers.append(name, value)

    def copy(self, cls=None):
        ''' Returns a copy of self. '''
        cls = cls or BaseResponse
        assert issubclass(cls, BaseResponse)
        copy = cls(status = self.status, headers = self.headers.copy().dict)
        if self._cookies:
            copy._cookies = SimpleCookie()
            copy._cookies.load(self._cookies.output(header=''))
        return copy

    def __iter__(self):
        return iter(self.body)

    def close(self):
        close = getattr(self.body, 'close', None)
        if close:
            close()

    @property
    def status_line(self):
        ''' The HTTP status line as a string (e.g. ``404 Not Found``).'''
        return self._status_line

    @property
    def status_code(self):
        ''' The HTTP status code as an integer (e.g. 404).'''
        return self._status_code

    @property
    def status(self):
        """ A writeable property to change the HTTP response status.

            It accepts either a numeric code (100-999) or a string with
            a custom reason phrase (e.g. "404 Brain not found").
            Both :data:`status_line` and :data:`status_code` are updated
            accordingly. The return value is always a status string.

        """
        return self._status_line

    @status.setter
    def status(self, status):
        if isinstance(status, int):
            code, status = status, _HTTP_STATUS_LINES.get(status)
        elif ' ' in status:
            status = status.strip()
            code = int(status.split()[0])
        else:
            raise ValueError('String status line without a reason phrase.')
        if not 100 <= code <= 999:
            raise ValueError('Status code out of range.')
        self._status_code = code
        self._status_line = str(status or ('%d Unknown' % code))

    @property
    def headerlist(self):
        """ WSGI conform list of (header, value) tuples. """
        headers = self._headers.items()
        bad_headers = self.bad_headers.get(self._status_code)
        if bad_headers:
            headers = (h for h in headers if h[0] not in bad_headers)
            need_ctype = False
        else:
            need_ctype = 'Content-Type' not in self._headers
        out = [
            (name, val.encode('utf8').decode('latin1'))
            for (name, vals) in headers
            for val in (vals if isinstance(vals, list) else [vals])
        ]
        if need_ctype:
            out.append(('Content-Type', self.default_content_type))
        if self._cookies:
            for c in self._cookies.values():
                out.append(
                    ('Set-Cookie', c.OutputString().encode('utf8').decode('latin1'))
                )
        return out

    content_type = HeaderProperty('Content-Type')
    content_length = HeaderProperty('Content-Length', reader=int)
    expires = HeaderProperty(
        'Expires',
        reader=lambda x: datetime.utcfromtimestamp(parse_date(x)),
        writer=lambda x: http_date(x)
    )

    @property
    def charset(self, default='UTF-8'):
        """ Return the charset specified in the content-type header (default: utf8). """
        if 'charset=' in self.content_type:
            return self.content_type.split('charset=')[-1].split(';')[0].strip()
        return default

    def set_cookie(self, name, value, secret=None, **options):
        ''' Create a new cookie or replace an old one. If the `secret` parameter is
            set, create a `Signed Cookie` (described below).

            :param name: the name of the cookie.
            :param value: the value of the cookie.
            :param secret: a signature key required for signed cookies.

            Additionally, this method accepts all RFC 2109 attributes that are
            supported by :class:`cookie.Morsel`, including:

            :param max_age: maximum age in seconds. (default: None)
            :param expires: a datetime object or UNIX timestamp. (default: None)
            :param domain: the domain that is allowed to read the cookie.
              (default: current domain)
            :param path: limits the cookie to a given path (default: current path)
            :param secure: limit the cookie to HTTPS connections (default: off).
            :param httponly: prevents client-side javascript to read this cookie
              (default: off, requires Python 2.6 or newer).

            If neither `expires` nor `max_age` is set (default), the cookie will
            expire at the end of the browser session (as soon as the browser
            window is closed).

            Signed cookies may store any pickle-able object and are
            cryptographically signed to prevent manipulation. Keep in mind that
            cookies are limited to 4kb in most browsers.

            Warning: Signed cookies are not encrypted (the client can still see
            the content) and not copy-protected (the client can restore an old
            cookie). The main intention is to make pickling and unpickling
            save, not to store secret information at client side.
        '''
        if not self._cookies:
            self._cookies = SimpleCookie()

        if secret:
            value = touni(cookie_encode((name, value), secret))
        elif not isinstance(value, str):
            raise TypeError('Secret key missing for non-string Cookie.')

        if len(value) > 4096:
            raise ValueError('Cookie value to long.')

        self._cookies[name] = value

        for key, value in options.items():
            if key == 'max_age':
                if isinstance(value, timedelta):
                    value = value.seconds + value.days * 24 * 3600
            if key == 'expires':
                value = http_date(value)
            self._cookies[name][key.replace('_', '-')] = value

    def delete_cookie(self, key, **kwargs):
        ''' Delete a cookie. Be sure to use the same `domain` and `path`
            settings as used to create the cookie. '''
        kwargs['max_age'] = -1
        kwargs['expires'] = 0
        self.set_cookie(key, '', **kwargs)

    def __repr__(self):
        out = []
        for name, value in self.headerlist:
            out.append('%s: %s' % (name.title(), value.strip()))
        return '\n'.join(out)


@ts_props(
    '_status_line', '_status_code',
    '_headers', '_cookies', 'body',
    store_name='_ts_props'
)
class Response(BaseResponse):
    __slots__ = _response_slots


###############################################################################
# Exceptions & Commons ########################################################
###############################################################################
class HTTPResponse(BaseResponse, OmbottException):

    __slots__ = _response_slots

    def __init__(self, body='', status=None, headers=None, **more_headers):
        super().__init__(body, status, headers, **more_headers)

    def apply(self, response):
        response._status_code = self._status_code
        response._status_line = self._status_line
        response._headers.clear()
        response._headers.update(self._headers)
        if self._cookies:
            response._cookies = self._cookies
        response.body = self.body


class HTTPError(HTTPResponse):
    __slots__ = ('exception', 'traceback')
    default_status = 500

    def __init__(self, status=None, body=None, exception=None, traceback=None,
                 **options):
        self.exception = exception
        self.traceback = traceback
        super().__init__(body, status, **options)


#: A dict to map HTTP status codes (e.g. 404) to phrases (e.g. 'Not Found')
HTTP_CODES = httplib.responses
HTTP_CODES[418] = "I'm a teapot"  # RFC 2324
HTTP_CODES[422] = "Unprocessable Entity"  # RFC 4918
HTTP_CODES[428] = "Precondition Required"
HTTP_CODES[429] = "Too Many Requests"
HTTP_CODES[431] = "Request Header Fields Too Large"
HTTP_CODES[511] = "Network Authentication Required"
_HTTP_STATUS_LINES = {code: f'{code} {msg}' for (code, msg) in HTTP_CODES.items()}



import base64
import cgi
from functools import partial
from http.cookies import SimpleCookie
from urllib.parse import (
    quote as urlquote,
    unquote as urlunquote,
    urljoin,
    SplitResult as UrlSplitResult
)

from ..common_helpers import (
    touni, tob,
    cookie_decode,
)
from .helpers import (
    cache_in,
    FormsDict,
    WSGIHeaderDict,
)


urlunquote = partial(urlunquote, encoding='latin1')


def cgi_monkey_patch():
    """Fix bug cgi.FieldStorage context manager bug https://github.com/python/cpython/pull/14815

    """
    def patch_exit(self, *exc):
        if self.file is not None:
            self.file.close()
    cgi.FieldStorage.__exit__ = patch_exit


cgi_monkey_patch()


class PropsMixin:

    @cache_in('environ[ ombott.app ]', read_only=True)
    def app(self):
        raise RuntimeError('This request is not connected to an application.')

    @cache_in('environ[ ombott.route ]', read_only=True)
    def route(self):
        """ The ombott :class:`Route` object that matches this request. """
        raise RuntimeError('This request is not connected to a route.')

    @cache_in('environ[ route.url_args ]', read_only=True)
    def url_args(self):
        """ The arguments extracted from the URL. """
        raise RuntimeError('This request is not connected to a route.')

    @property
    def path(self):
        ''' The value of ``PATH_INFO`` with exactly one prefixed slash (to fix
            broken clients and avoid the "empty path" edge case). '''
        return '/' + self._env_get('PATH_INFO', '').lstrip('/')

    @property
    def method(self):
        ''' The ``REQUEST_METHOD`` value as an uppercase string. '''
        return self._env_get('REQUEST_METHOD', 'GET').upper()

    @cache_in('environ[ ombott.request.headers ]', read_only=True)
    def headers(self):
        ''' A :class:`WSGIHeaderDict` that provides case-insensitive access to
            HTTP request headers. '''
        return WSGIHeaderDict(self.environ)

    @cache_in('environ[ ombott.request.cookies ]', read_only=True)
    def cookies(self):
        """ Cookies parsed into a :class:`FormsDict`. Signed cookies are NOT
            decoded. Use :meth:`get_cookie` if you expect signed cookies. """
        cookies = SimpleCookie(self._env_get('HTTP_COOKIE', '')).values()
        return FormsDict((c.key, c.value) for c in cookies)

    def get_cookie(self, key, default=None, secret=None):
        """ Return the content of a cookie. To read a `Signed Cookie`, the
            `secret` must match the one used to create the cookie (see
            :meth:`BaseResponse.set_cookie`). If anything goes wrong (missing
            cookie or wrong signature), return a default value. """
        value = self.cookies.get(key)
        if secret and value:
            dec = cookie_decode(value, secret)  # (key, value) tuple or None
            return dec[1] if dec and dec[0] == key else default
        return value or default

    @cache_in('environ[ ombott.request.params ]', read_only=True)
    def params(self):
        """ A :class:`FormsDict` with the combined values of :attr:`query` and
            :attr:`forms`. File uploads are stored in :attr:`files`. """
        return FormsDict(self.query, **self.forms)

    @cache_in('environ[ ombott.request.url ]', read_only=True)
    def url(self):
        """ The full request URI including hostname and scheme. If your app
            lives behind a reverse proxy or load balancer and you get confusing
            results, make sure that the ``X-Forwarded-Host`` header is set
            correctly. """
        return self.urlparts.geturl()

    @cache_in('environ[ ombott.request.urlparts ]', read_only=True)
    def urlparts(self):
        ''' The :attr:`url` string as an :class:`urlparse.SplitResult` tuple.
            The tuple contains (scheme, host, path, query_string and fragment),
            but the fragment is always empty because it is not visible to the
            server. '''
        env_get = self._env_get
        http = env_get('HTTP_X_FORWARDED_PROTO') or env_get('wsgi.url_scheme', 'http')
        host = env_get('HTTP_X_FORWARDED_HOST') or env_get('HTTP_HOST')
        if not host:
            # HTTP 1.1 requires a Host-header. This is for HTTP/1.0 clients.
            host = env_get('SERVER_NAME', '127.0.0.1')
            port = env_get('SERVER_PORT')
            if port and port != ('80' if http == 'http' else '443'):
                host += ':' + port
        path = urlquote(self.fullpath)
        return UrlSplitResult(http, host, path, env_get('QUERY_STRING'), '')

    @cache_in('environ[ ombott.request.fullpath ]', read_only=True)
    def fullpath(self):
        """ Request path including :attr:`script_name` (if present). """
        appname = self._env_get(self.config.app_name_header, '/')
        return urljoin(self.script_name, self.path[len(appname):].lstrip('/'))

    @property
    def query_string(self):
        """ The raw :attr:`query` part of the URL (everything in between ``?``
            and ``#``) as a string. """
        return self._env_get('QUERY_STRING', '')

    @cache_in('environ[ ombott.request.script_name ]', read_only=True)
    def script_name(self):
        ''' The initial portion of the URL's `path` that was removed by a higher
            level (server or routing middleware) before the application was
            called. This script path is returned with leading and tailing
            slashes. '''
        env_get = self._env_get
        script_name = env_get('SCRIPT_NAME')
        if not script_name and self.config.allow_x_script_name:
            script_name = env_get('HTTP_X_SCRIPT_NAME')
        return '/' + script_name.strip('/') + '/' if script_name else '/'

    @property
    def is_xhr(self):
        ''' True if the request was triggered by a XMLHttpRequest. This only
            works with JavaScript libraries that support the `X-Requested-With`
            header (most of the popular libraries do). '''
        requested_with = self._env_get('HTTP_X_REQUESTED_WITH', '')
        return requested_with.lower() == 'xmlhttprequest'

    @property
    def is_ajax(self):
        ''' Alias for :attr:`is_xhr`. "Ajax" is not the right term. '''
        return self.is_xhr

    @cache_in('environ[ ombott.request.is_json_requested ]', read_only=True)
    def is_json_requested(self):
        accept = self._env_get('HTTP_ACCEPT')
        if accept:
            return accept.startswith('application/json')

    @property
    def auth(self):
        """ HTTP authentication data as a (user, password) tuple. This
            implementation currently supports basic (not digest) authentication
            only. If the authentication happened at a higher level (e.g. in the
            front web-server or a middleware), the password field is None, but
            the user field is looked up from the ``REMOTE_USER`` environ
            variable. On any errors, None is returned. """

        def parse_auth(header):
            """ Parse rfc2617 HTTP authentication header string (basic) and return (user,pass) tuple or None"""
            try:
                method, data = header.split(None, 1)
                if method.lower() == 'basic':
                    user, pwd = touni(base64.b64decode(tob(data))).split(':', 1)
                    return user, pwd
            except (KeyError, ValueError):
                return None

        basic = parse_auth(self._env_get('HTTP_AUTHORIZATION', ''))
        if basic:
            return basic
        ruser = self._env_get('REMOTE_USER')
        if ruser:
            return (ruser, None)
        return None

    @cache_in('environ[ ombott.request.remote_route ]', read_only=True)
    def remote_route(self):
        """ A list of all IPs that were involved in this request, starting with
            the client IP and followed by zero or more proxies. This does only
            work if all proxies support the ```X-Forwarded-For`` header. Note
            that this information can be forged by malicious clients. """
        proxy = self._env_get('HTTP_X_FORWARDED_FOR')
        if proxy:
            return [ip.strip() for ip in proxy.split(',')]
        remote = self._env_get('REMOTE_ADDR')
        return [remote] if remote else []

    @property
    def remote_addr(self):
        """ The client IP as a string. Note that this information can be forged
            by malicious clients. """
        route = self.remote_route
        return route[0] if route else None



import sys
import json
import functools
from traceback import format_exc
import itertools
from urllib.parse import urljoin

from .common_helpers import (
    html_escape,
    tob,
    cached_property,
    WSGIFileWrapper,
    SimpleConfig,
)
from .router import HookTypes, RadiRouter
from .request import Request, errors as request_errors
from .response import Response, HTTPResponse, HTTPError
from . import server_adapters
from . import error_render

__version__ = "0.0.14"

HTTP_METHODS = 'DELETE GET HEAD OPTIONS PATCH POST PUT'.split()


@SimpleConfig.keys_holder
class DefaultConfig(SimpleConfig):
    catchall = True
    debug = False
    domain_map = {}

    # request specific
    app_name_header = ''
    errors_map = {
        request_errors.RequestError: HTTPError(400, 'Bad request'),
        request_errors.BodySizeError: HTTPError(413, 'Request entity too large'),
        request_errors.BodyParsingError: HTTPError(400, 'Error while parsing chunked transfer body'),
    }
    max_body_size = None
    max_memfile_size = 100 * 1024
    allow_x_script_name = False


class _closeiter:
    ''' This only exists to be able to attach a .close method to iterators that
        do not support attribute assignment (most of itertools). '''

    def __init__(self, iterator, close=None):
        self.iterator = iterator
        self.close_callbacks = close if isinstance(close, (list, tuple)) else [close]

    def __iter__(self):
        return iter(self.iterator)

    def close(self):
        [cb() for cb in self.close_callbacks]


def run(app=None, server='wsgiref', host='127.0.0.1', port=8080,
        quiet=False, **kwargs):
    _stderr = sys.stderr.write
    try:
        app = app or default_app()
        if not callable(app):
            raise ValueError("Application is not callable: %r" % app)
        server_names = server_adapters.server_names
        if server in server_names:
            server = server_names.get(server)
        server = server(host=host, port=port, **kwargs)
        server.quiet = server.quiet or quiet
        if not server.quiet:
            _stderr("Ombott v%s server starting up (using %s)...\n" % (__version__, repr(server)))
            _stderr(f"Listening on http{'s' if kwargs.get('certfile', None) else ''}://{server.host}:{server.port}/\n")
            _stderr("Hit Ctrl-C to quit.\n\n")
        server.run(app)
    except KeyboardInterrupt:
        pass
    except (SystemExit, MemoryError):
        raise
    except:  # noqa
        raise


def with_method_shortcuts(methods):
    def injector(cls):
        for m in methods:
            setattr(cls, m.lower(), functools.partialmethod(cls.route, method=m))
        return cls
    return injector

###############################################################################
# Application Object ###########################################################
###############################################################################


@with_method_shortcuts(HTTP_METHODS)
class Ombott:
    __slots__ = ('config', 'router', 'request', 'response', '_route_hooks', 'error_handlers', '__dict__')

    def __init__(self, config=None):
        self.config = config = DefaultConfig(config)
        self.router = RadiRouter()
        self.request = Request(config=config)
        self.response = Response()
        self._route_hooks = {}
        self.error_handlers = {'404-hooks': {}}

    def setup(self, config=None):
        self.config = config = DefaultConfig(config)
        self.request.setup(config)

    def run(self, **kwargs):
        ''' Calls :func:`run` with the same parameters. '''
        run(self, **kwargs)

    def to_route(self, path, verb):
        if verb == 'HEAD':
            methods = [verb, 'GET', 'ANY']
        else:
            methods = [verb, 'ANY']
        end_point, error404_405 = self.router.resolve(path, methods)
        return (end_point, error404_405)

    def add_route(self, rule, method, handler, name=None, *, overwrite=False):
        return self.router.add(rule, method, handler, name, overwrite=overwrite)

    def remove_route(self, rule=None, *, route_pattern=None, name=None):
        self.router.remove(rule, route_pattern=route_pattern, name=name)

    def route(self, rule=None, method='GET', callback=None,
              *, name=None, overwrite=False):
        def decorator(callback):
            self.add_route(rule, method, callback, name, overwrite=overwrite)
            return callback
        return decorator(callback) if callback else decorator

    # overwrites by @with_method_shortcuts()
    def delete(self, rule=None, callback=None, *, name=None, overwrite=False): pass
    def get(self, rule=None, callback=None, *, name=None, overwrite=False): pass
    def head(self, rule=None, callback=None, *, name=None, overwrite=False): pass
    def options(self, rule=None, callback=None, *, name=None, overwrite=False): pass
    def patch(self, rule=None, callback=None, *, name=None, overwrite=False): pass
    def post(self, rule=None, callback=None, *, name=None, overwrite=False): pass
    def put(self, rule=None, callback=None, *, name=None, overwrite=False): pass

    @property
    def routes(self):
        return self.router.routes

    __hook_names = ('before_request', 'after_request')
    __hook_reversed = {'after_request'}

    @cached_property
    def _hooks(self):
        return {name: [] for name in self.__hook_names}

    def add_hook(self, name, func):
        """Attach a callback to a hook.

        Three hooks are currently implemented:
            `before_request`
                Executed once before each request. The request context is
                available, but no routing has happened yet.
            `after_request`
                Executed once after each request regardless of its outcome.

        """
        if name in self.__hook_reversed:
            self._hooks[name].insert(0, func)
        else:
            self._hooks[name].append(func)

    def on(self, name, func=None):
        if not func:  # used as decorator
            def decorator(func):
                self.add_hook(name, func)
                return func
            return decorator
        else:
            self.add_hook(name, func)

    def remove_hook(self, name, func):
        if func in self._hooks[name]:
            self._hooks[name].remove(func)
            return True

    def emit(self, name, *args, **kwargs):
        [hook(*args, **kwargs) for hook in self._hooks[name][:]]

    def on_route(self, rule, func=None):
        if not func:  # used as decorator
            def decorator(func):
                self.router.add_hook(rule, func)
                return func
            return decorator
        else:
            self.router.add_hook(rule, func)

    def remove_route_hook(self, rule):
        self.router.remove_hook(rule)

    def error(self, code=500, rule=None):
        """ Decorator: Register an output handler for a HTTP error code"""
        code = int(code)

        def wrapper(handler):
            if code == 404 and rule:
                route_pattern = self.router.add_hook(
                    rule, handler, hook_type=HookTypes.PARTIAL
                )
                self.error_handlers['404-hooks'][route_pattern] = handler
            else:
                self.error_handlers[code] = handler
            return handler
        return wrapper

    def default_error_handler(self, res):
        if self.request.is_json_requested:
            ret = json.dumps(dict(
                body = res.body,
                exception = repr(res.exception),
                traceback = res.traceback
            ))
            self.response.headers['Content-Type'] = 'application/json'
        else:
            ret = error_render.render(res, self.request.url, self.config.debug)
        return ret

    @staticmethod
    def handler(app: 'Ombott', route, kwargs, route_hooks, error404_405):
        if error404_405:
            status, body, extra = error404_405
            if status == 405:
                raise HTTPError(status, body, Allow=extra)
            else:  # not found
                hooks_collected = extra['hooks']
                if hooks_collected:
                    route_pos, hooks = hooks_collected[-1]
                    partial_hook = hooks[HookTypes.PARTIAL]
                    if partial_hook:
                        hook_route = app.request.path[:1 + route_pos]
                        return partial_hook(hook_route, extra['param_values'])
                raise HTTPError(status, body)

        if route_hooks:
            path = app.request.path
            for route_pos, hooks in route_hooks:
                hook = hooks[HookTypes.SIMPLE]
                if hook:
                    hook(path[:1 + route_pos])
        return route(**kwargs)

    def _handle(self, environ):
        response = self.response
        request = self.request

        path = environ['ombott.raw_path'] = environ['PATH_INFO']
        try:
            path = path.encode('latin1').decode('utf8')
        except UnicodeError:
            return HTTPError(400, 'Invalid path string. Expected UTF-8')
        environ['PATH_INFO'] = path
        try:  # init thread
            environ['ombott.app'] = self
            request.__init__(environ)
            response.__init__()
            try:  # routing
                self.emit('before_request')
                route, kwargs, route_hooks = (None, None, None)
                end_point, error404_405 = self.to_route(request.path, request.method)
                if end_point:
                    route, kwargs, route_hooks = end_point
                    environ['ombott.route'] = route
                    environ['route.url_args'] = kwargs
                    environ['route.hooks'] = route_hooks
                return self.handler(self, route, kwargs, route_hooks, error404_405)
            finally:
                self.emit('after_request')
        except HTTPResponse as resp:
            return resp
        except (KeyboardInterrupt, SystemExit, MemoryError):
            raise
        except Exception as err500:
            # raise
            stacktrace = format_exc()
            environ['wsgi.errors'].write(stacktrace)
            return HTTPError(500, "Internal Server Error", err500, stacktrace)

    def _cast(self, out):
        """ Try to convert the parameter into something WSGI compatible and set
        correct HTTP headers when possible.
        Support: False, str, unicode, dict, HTTPResponse, HTTPError, file-like,
        iterable of strings and iterable of unicodes
        """

        response = self.response
        resp_headers = response.headers
        request = self.request
        loops_cnt = 0
        while True:   # <-------
            loops_cnt += 1
            if loops_cnt > 1000:
                out = HTTPError(500, 'too many iterations')
                out.apply(response)
                out = self.default_error_handler(out)

            # Empty output is done here
            if not out:
                resp_headers.setdefault('Content-Length', 0)
                return []

            if isinstance(out, str):
                out = out.encode(response.charset)

            # Byte Strings are just returned
            if isinstance(out, bytes):
                resp_headers.setdefault('Content-Length', len(out))
                return [out]

            if isinstance(out, HTTPError):
                out.apply(response)
                out = self.error_handlers.get(
                    out.status_code,
                    self.default_error_handler
                )(out); continue                         # -----------------^

            if isinstance(out, HTTPResponse):
                out.apply(response)
                out = out.body; continue                 # -----------------^

            # File-like objects.
            if hasattr(out, 'read'):
                if 'wsgi.file_wrapper' in request.environ:
                    return request.environ['wsgi.file_wrapper'](out)
                elif hasattr(out, 'close') or not hasattr(out, '__iter__'):
                    return WSGIFileWrapper(out)

            # Handle Iterables. We peek into them to detect their inner type.
            try:
                iout = iter(out)
                first = next(iout)
                while not first:
                    first = next(iout)
            except StopIteration:
                out = ''; continue                               # -----------------^
            except HTTPResponse as rs:
                first = rs
            except (KeyboardInterrupt, SystemExit, MemoryError):
                raise
            except Exception as err500:
                # if not self.catchall: raise
                first = HTTPError(500, 'Unhandled exception', err500, format_exc())

            # These are the inner types allowed in iterator or generator objects.
            if isinstance(first, HTTPResponse):
                out = first; continue                            # -----------------^
            elif isinstance(first, bytes):
                new_iter = itertools.chain([first], iout)
            elif isinstance(first, str):
                new_iter = (it.encode(response.charset) for it in itertools.chain([first], iout))
            else:
                out = HTTPError(500, f'Unsupported response type: {type(first)}')
                continue                                         # -----------------^
            close = getattr(out, 'close', None)
            if close:
                new_iter = _closeiter(new_iter, close)
            return new_iter

    def wsgi(self, environ, start_response):
        config: DefaultConfig = self.config
        response = self.response

        domain_map = config.domain_map
        if domain_map:
            app_name = domain_map(environ.get('HTTP_X_FORWARDED_HOST') or environ.get('HTTP_HOST'))
            if app_name:
                environ[config.app_name_header] = '/' + app_name
                environ["PATH_INFO"] = '/' + app_name + environ["PATH_INFO"]

        try:
            out = self._cast(self._handle(environ))
            # rfc2616 section 4.3
            if (
                response._status_code in {100, 101, 204, 304}
                or environ['REQUEST_METHOD'] == 'HEAD'
            ):
                close = getattr(out, 'close', None)
                if close:
                    close()
                out = []
            start_response(response._status_line, response.headerlist)
            return out
        except (KeyboardInterrupt, SystemExit, MemoryError):
            raise
        except Exception as _e:
            if not self.config.catchall:
                raise

            err = '<h1>Critical error while processing request: %s</h1>' \
                  % html_escape(environ.get('PATH_INFO', '/'))
            if self.config.debug:
                err += (
                    '<h2>Error:</h2>\n<pre>\n%s\n</pre>\n'
                    '<h2>Traceback:</h2>\n<pre>\n%s\n</pre>\n'
                    % (html_escape(repr(_e)), html_escape(format_exc()))
                )
            environ['wsgi.errors'].write(err)
            headers = [('Content-Type', 'text/html; charset=UTF-8')]
            start_response('500 INTERNAL SERVER ERROR', headers, sys.exc_info())
            return [tob(err)]

    def __call__(self, environ, start_response):
        return self.wsgi(environ, start_response)


###############################################################################
# Application Helper ###########################################################
###############################################################################


def abort(code=500, text='Unknown Error'):
    """ Aborts execution and causes a HTTP error. """
    raise HTTPError(code, text)


def redirect(location, code=None):
    """ Aborts execution and causes a 303 or 302 redirect, depending on
        the HTTP protocol version.

    """
    response = Globals.response
    request = Globals.request
    url = location
    if not code:
        code = 303 if request.get('SERVER_PROTOCOL') == "HTTP/1.1" else 302
    res = response.copy(cls=HTTPResponse)
    res.status = code
    res.body = ""
    res.headers['Location'] = urljoin(request.url, url)
    raise res


class Globals:
    app = Ombott()
    route = app.route
    on_route = app.on_route
    request = app.request
    response = app.response
    error = app.error


def default_app():
    return Globals.app



import re


class _RouteFilterExhaust:
    __slots__ = ('value', 'move_len', 'selector')

    def __init__(self, value, move_len, selector = None):
        self.value = value
        self.move_len = move_len
        self.selector = selector

    def get(self):
        return self.value, self.move_len, self.selector


def _rex(conf):
    def f_in(matched, re_obj):
        for selector, v in enumerate(re_obj.groups()):
            if v is not None:
                selector += 1
                break
        else:
            selector = None
            v = matched
        return _RouteFilterExhaust(v, re_obj.end(), selector)
    return conf, f_in, None


class FilterFactory:
    filters = {
        're':    lambda conf: (conf, None, None),
        'rex':   _rex,
        'int':   lambda conf: (r'-?\d+', int, lambda x: str(int(x))),
        'float': lambda conf: (r'-?\d+(\.\d+)?', float, lambda x: str(float(x))),
        'path':  lambda conf: (f'.+(?={re.escape(conf)})' if conf else '.+$', None, None)
    }
    _filter_cache = {}

    @classmethod
    def make_filter(cls, filter: str, args: str):
        if not filter:
            return None, None
        fkey = f'{filter}({args})'

        handler_f_out = cls._filter_cache.get(fkey)
        if handler_f_out:
            return handler_f_out

        mask, f_in, f_out = cls.filters[filter](args)
        mask = re.compile(mask)
        if f_in:
            code = getattr(f_in, '__code__', None)
            if code and code.co_argcount != 1:  # 0 means f_in(*args)
                def handler(param):
                    selector = None
                    tmp = mask.match(param)
                    if not tmp:
                        return None, 0, selector
                    ret = f_in(tmp.group(), tmp)
                    if isinstance(ret, _RouteFilterExhaust):
                        ret, pos, selector = ret.get()
                    else:
                        pos = tmp.end()
                    return ret, pos, selector
            else:
                def handler(param):
                    tmp = mask.match(param)
                    if not tmp:
                        return None, 0, None
                    return f_in(tmp.group()), tmp.end(), None
        else:
            def handler(param):
                tmp = mask.match(param)
                if not tmp:
                    return None, 0, None
                return tmp.group(), tmp.end(), None
        cls._filter_cache[fkey] = [handler, f_out]
        return handler, f_out



from .ombott import ( # noqa
    Ombott,
    DefaultConfig,
    HTTPResponse,
    HTTPError,
    Response,
    Request,
    abort,
    default_app,
    redirect,

    Globals as _Globals,
    default_app,  # for bottle compat
    run,
    __version__
)

route = _Globals.route
on_route = _Globals.on_route
request = _Globals.request
response = _Globals.response
error = _Globals.error
app = _Globals.app

from .common_helpers import SimpleConfig    # noqa
from .static_stream import static_file        # noqa

__author__ = "Valery Kucherov <valq7711@gmail.com>"
__copyright__ = "Copyright (c) 2009-2012, Marcel Hellkamp; Copyright (C) 2021-2022 Valery Kucherov"
__license__ = "MIT"

__all__ = ('__version__', '__license__', '__copyright__', '__author__')


import json as json_mod
import cgi
from tempfile import TemporaryFile
from io import BytesIO
from functools import partial

from ..common_helpers import touni

from .helpers import (
    parse_qsl,
    cache_in,
    FileUpload,
    FormsDict
)
from .errors import RequestError, BodyParsingError, BodySizeError


# fix bug cgi.FieldStorage context manager bug https://github.com/python/cpython/pull/14815
def _cgi_monkey_patch():
    def patch_exit(self, *exc):
        if self.file is not None:
            self.file.close()
    cgi.FieldStorage.__exit__ = patch_exit


_cgi_monkey_patch()


def _iter_body(read, buff_size, *, content_length):
    rest_len = content_length
    while rest_len > 0:
        part_size = min(rest_len, buff_size)
        part = read(part_size)
        if not part:
            break
        yield part
        rest_len -= part_size


def _iter_chunked(read, buff_size):
    r, n, rn, sem = b'\r', b'\n', b'\r\n', b';'
    header_size_buff = []
    parsing_err = BodyParsingError()
    while True:

        # read header to get chunk size
        header_size_buff.clear()
        read_len = 0
        seen_r = False
        seen_sem = False
        while True:   # header[-2:] != rn:
            c = read(1)
            read_len += 1
            if not c or read_len > buff_size:
                raise parsing_err
            # catch `\r\n`
            if seen_r and c == n:
                break
            seen_r = c == r
            # maybe behind `;` (chunk extra info)
            if seen_sem:
                # continue reading until `\r\n`
                continue
            seen_sem = c == sem
            if seen_r or seen_sem:
                continue
            header_size_buff.append(c)

        chunk_size = b''.join(header_size_buff)
        try:
            rest_len = int(chunk_size.strip(), 16)
        except ValueError:
            raise parsing_err
        if rest_len == 0:
            break

        # read chunk body
        while rest_len > 0:
            part_size = min(rest_len, buff_size)
            part = read(part_size)
            if not part:
                raise parsing_err
            yield part
            rest_len -= part_size
        if read(2) != rn:
            raise parsing_err


def _body_read(read, buff_size, *, content_length = None, chunked = None, max_body_size = None):
    body_iter = (
        _iter_chunked if chunked
        else partial(_iter_body, content_length = content_length)
    )
    body, body_size, is_temp_file = BytesIO(), 0, False
    for part in body_iter(read, buff_size):
        body.write(part)
        body_size += len(part)
        if max_body_size is not None and body_size > max_body_size:
            raise BodySizeError()
        if not is_temp_file and body_size > buff_size:
            body, tmp = TemporaryFile(mode='w+b'), body
            body.write(tmp.getvalue())
            del tmp
            is_temp_file = True
    return body


class BodyMixin:

    @cache_in('environ[ ombott.request.content_length ]', read_only=True)
    def content_length(self):
        ''' The request body length as an integer. The client is responsible to
            set this header. Otherwise, the real length of the body is unknown
            and -1 is returned. In this case, :attr:`body` will be empty. '''
        return int(self.environ.get('CONTENT_LENGTH') or -1)

    @cache_in('environ[ ombott.request.content_type ]', read_only=True)
    def content_type(self):
        ''' The Content-Type header as a lowercase-string (default: empty). '''
        return self.environ.get('CONTENT_TYPE', '').lower()

    @cache_in('environ[ ombott.request.ctype ]', read_only=True)
    def ctype(self):
        ctype = self.content_type.split(';')
        return [t.strip() for t in ctype]

    @property
    def chunked(self):
        ''' True if Chunked transfer encoding was. '''
        return 'chunked' in self.environ.get('HTTP_TRANSFER_ENCODING', '').lower()

    @property
    def body(self):
        ret = self._body
        ret.seek(0)
        return ret

    @cache_in('environ[ ombott.request.query ]', read_only=True)
    def query(self):
        ''' The :attr:`query_string` parsed into a :class:`FormsDict`. These
            values are sometimes called "URL arguments" or "GET parameters", but
            not to be confused with "URL wildcards" as they are provided by the
            :class:`Router`. '''
        ret = FormsDict()
        qs = self._env_get('QUERY_STRING', '')
        if qs:
            parse_qsl(qs, setitem = ret.__setitem__)
        self.environ['ombott.request.get'] = ret
        return ret

    #: An alias for :attr:`query`.
    GET = query

    @cache_in('environ[ ombott.request.json ]', read_only=True)
    def json(self):
        ''' If the ``Content-Type`` header is ``application/json``, this
            property holds the parsed content of the request body. Only requests
            smaller than :attr:`MEMFILE_MAX` are processed to avoid memory
            exhaustion. '''
        if self.ctype[0] == 'application/json':
            b = self._get_body_string()
            if not b:
                return None
            return json_mod.loads(b)
        return None

    @cache_in('environ[ ombott.request.post ]', read_only=True)
    def POST(self):
        """ The values of :attr:`forms` and :attr:`files` combined into a single
            :class:`FormsDict`. Values are either strings (form values) or
            instances of :class:`cgi.FieldStorage` (file uploads).

        """
        env = self.environ
        files = env['ombott.request.files'] = FormsDict()
        post = FormsDict()

        # We default to application/x-www-form-urlencoded for everything that
        # is not multipart and take the fast path
        ctype = self.content_type
        if not ctype.startswith('multipart/'):
            if ctype.startswith('application/json'):
                post.update(self.json)
            else:
                parse_qsl(
                    touni(self._get_body_string(), 'latin1'),
                    setitem = post.__setitem__
                )
            env['ombott.request.forms'] = post
            return post

        forms = env['ombott.request.forms'] = FormsDict()

        safe_env = {'QUERY_STRING': ''}  # Build a safe environment for cgi
        for key in ('REQUEST_METHOD', 'CONTENT_TYPE', 'CONTENT_LENGTH'):
            if key in env:
                safe_env[key] = env[key]
        args = dict(
            fp=self.body,
            environ=safe_env,
            keep_blank_values=True,
            encoding='utf8'
        )
        listified = set()
        with cgi.FieldStorage(**args) as data:
            self['_cgi.FieldStorage'] = data  # http://bugs.python.org/issue18394#msg207958
            data = data.list or []
            for item in data:
                if item.filename:
                    it = FileUpload(
                        item.file, item.name,
                        item.filename, item.headers
                    )
                    dct = files
                else:
                    it = item.value
                    dct = forms
                key = item.name

                if key in post:
                    el = post[key]
                    if key not in listified:
                        el = post[key] = dct[key] = [el]
                        listified.add(key)
                    el.append(it)
                else:
                    post[key] = dct[key] = it
        return post

    @cache_in('environ[ ombott.request.forms ]', read_only=True)
    def forms(self):
        """ Form values parsed from an `url-encoded` or `multipart/form-data`
            encoded POST or PUT request body. The result is returned as a
            :class:`FormsDict`. All keys and values are strings. File uploads
            are stored separately in :attr:`files`.
        """
        self.POST
        return self.environ['ombott.request.forms']

    @cache_in('environ[ ombott.request.files ]', read_only=True)
    def files(self):
        """ File uploads parsed from `multipart/form-data` encoded POST or PUT
            request body. The values are instances of :class:`FileUpload`.

        """
        self.POST
        return self.environ['ombott.request.files']

    @cache_in('environ[ ombott.request.body ]', read_only=True)
    def _body(self):
        try:
            body = _body_read(
                self.environ['wsgi.input'].read,
                self.config.max_memfile_size,
                content_length = self.content_length,
                chunked = self.chunked,
                max_body_size = self.config.max_body_size
            )
        except RequestError as err:
            self._raise(err, RequestError)
        self.environ['wsgi.input'] = body
        body.seek(0)
        return body

    def _get_body_string(self):
        ''' read body until content-length or MEMFILE_MAX into a string. Raise
            BodySizeError on requests that are to large.

        '''
        self._body.seek(0)
        read = self._body.read
        max_content_length = self.config.max_memfile_size
        content_length = self.content_length

        if content_length > max_content_length:
            raise self._raise(BodySizeError(), RequestError)
        if content_length < 0:
            content_length = max_content_length + 1
        data = read(content_length)
        if len(data) > max_content_length:  # Fail fast
            raise self._raise(BodySizeError(), RequestError)
            # raise HTTPError(413, 'Request to large')
        return data



from setuptools import setup, find_packages
import ombott

setup(
    name="ombott",
    version=ombott.__version__,
    url="https://github.com/valq7711/ombott",
    license=ombott.__license__,
    author=ombott.__author__,
    author_email="valq7711@gmail.com",
    maintainer=ombott.__author__,
    maintainer_email="valq7711@gmail.com",
    description="One More BOTTle",
    platforms="any",
    keywords='python webapplication',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires='>=3.7',
    packages=find_packages('.'),
    package_data={'ombott': ['error.html']}
)




from ..common_helpers import ts_props, SimpleConfig
from ..mixable import Mixable

from .props_mixin import PropsMixin
from .body_mixin import BodyMixin


class RequestConfig(SimpleConfig):
    app_name_header = ''
    errors_map = {}
    max_body_size = None
    max_memfile_size = 100 * 1024
    allow_x_script_name = False


class BaseRequest:
    __slots__ = ('environ', '_env_get', '__listeners__', 'config', '_ts_props')

    def __new__(cls, environ = None, *, config=None):
        self = super().__new__(cls)
        self.__listeners__ = {}
        self.on('env_changed', cls._on_env_changed)
        self.config = RequestConfig.get_from(config)
        return self

    def __init__(self, environ = None, *, config = None):
        # self.environ = some  - also does self._env_get = some.get
        self.environ = {} if environ is None else environ
        self.environ['ombott.request'] = self

    def setup(self, config):
        self.config = RequestConfig.get_from(config)

    def _raise(self, err, except_class = None):
        errors_map = self.config.errors_map
        for err_cls in (err.__class__, except_class):
            out_err = errors_map.get(err_cls)
            if out_err:
                err = out_err
                break
        raise err

    @staticmethod
    def _on_env_changed(request, key, v):
        todelete = ()
        if key == 'wsgi.input':
            todelete = ('forms', 'files', 'params', 'post', 'json', 'body')
        elif key == 'QUERY_STRING':
            todelete = ('query', 'params')
        elif key.startswith('HTTP_'):
            todelete = ('headers', 'cookies')
        env = request.environ
        [env.pop('ombott.request.' + key, None) for key in todelete]

    def on(self, e, cb):
        if e not in self.__listeners__:
            self.__listeners__[e] = []
        self.__listeners__[e].append(cb)
        return lambda: self.__listeners__[e].remove(cb)

    def off(self, e, cb):
        self.__listeners__[e].remove(cb)

    def emit(self, e, *a, **kw):
        if e not in self.__listeners__:
            return
        [cb(self, *a, **kw) for cb in self.__listeners__[e]]

    def copy(self):
        """ Return a new :class:`Request` with a shallow :attr:`environ` copy.

        NOTE: __listeners__ are not copied

        """
        copy = self.__class__(self.environ.copy(), config = self.config)
        return copy

    def get(self, value, default=None):
        return self._env_get(value, default)

    def keys(self):
        return self.environ.keys()

    def __iter__(self):
        return iter(self.environ)

    def __len__(self):
        return len(self.environ)

    def __getitem__(self, key):
        return self.environ[key]

    def __setitem__(self, key, value):
        """ Change an environ value and clear all caches that depend on it. """

        if self._env_get('ombott.request.readonly'):
            raise KeyError('The environ dictionary is read-only.')

        env = self.environ
        if key in env and env[key] in [value]:  # `in` performs 2 OR-gluing tests (`is` OR `==`)
            return

        env[key] = value
        self.emit('env_changed', key, value)

    def __delitem__(self, key):
        self[key] = ""
        del self.environ[key]

    def __getattr__(self, name):
        ''' Search in self.environ for additional user defined attributes. '''
        if name in self.__slots__:
            return
        try:
            var = self.environ['ombott.request.ext.%s' % name]
            getter = getattr(var, '__get__', None)
            return getter(self) if getter else var
        except KeyError:
            raise AttributeError('Attribute %r not defined.' % name)

    def __setattr__(self, name, value):
        if name in self.__slots__:
            ret = object.__setattr__(self, name, value)
            if name == 'environ':
                object.__setattr__(self, '_env_get', value.get)
            return ret
        self.environ['ombott.request.ext.%s' % name] = value
        return value

    def __repr__(self):
        return '<%s: %s %s>' % (self.__class__.__name__, self.method, self.url)


@ts_props('environ', '_env_get', store_name = '_ts_props')
class Request(Mixable, BaseRequest, PropsMixin, BodyMixin):
    _as_mixins = [PropsMixin, BodyMixin]



from enum import IntEnum
from .radidict import RadiDict, DATA, HOOKS
from .filter_factory import FilterFactory
from .errors import RouteBuildError, RouteMethodError
from .parser import Parser


class RouteMethod:
    __slots__ = ('route', 'name', 'handler', 'meta')

    def __init__(self, route, name, handler, meta=None):
        self.route = route
        self.name = name
        self.handler = handler
        self.meta = meta

    def remove(self):
        self.route.remove_method(self.name)

    @classmethod
    def get_func_fullname(cls, func):
        return f'{func.__module__}.{func.__qualname__}'

    @property
    def handler_fullname(self):
        return self.get_func_fullname(self.handler)

    def __str__(self):
        return '{}: {}'.format(self.name, self.handler)

    def __repr__(self):
        return '<{}:{} {}>'.format(self.route.rule, self.name, self.handler)

    def __call__(self, *a, **kw):
        return self.handler(*a, **kw)


class Route:
    __slots__ = (
        'rule', '_methods', 'pattern', 'params', 'filters',
        'pattern_out', 'filters_out'
    )

    parser = Parser()

    anon_prefix = 'anon-'

    make_filter = staticmethod(FilterFactory.make_filter)

    def __init__(self, rule):
        #: The path-rule string (e.g. ``/wiki/:page``).
        self.rule = rule
        #: The HTTP methods dict like {'GET': {'handler': <callable>, 'meta': <any>} }
        self._methods = dict()
        pattern, params, filters, pattern_out, filters_out = self.__class__.parse_rule(rule)
        self.pattern = pattern  # never starts from '/'
        self.params = params
        self.filters = filters

        self.pattern_out = pattern_out
        self.filters_out = filters_out

    def url(self, *args, **kw):

        params = self.params
        if not params:
            return self.pattern_out

        filters = self.filters
        filters_out = self.filters_out
        anon_prefix = self.anon_prefix
        pattern_out = self.pattern_out

        ret = []
        pidx = 0
        args_idx = 0
        cidx = 0
        clen = 0
        end = 0
        for c in pattern_out:
            if c != '\r':
                clen += 1
                continue
            end = cidx
            if clen:
                end += clen
                clen = 0
                ret.append(pattern_out[cidx:end])
            cidx = end + 1

            pname = params[pidx]
            f_out = filters_out[pidx]
            f_in = filters[pidx]
            pidx += 1

            if pname.startswith(anon_prefix):
                prt = args[args_idx]
                args_idx += 1
            else:
                prt = kw[pname]
            if f_out:
                prt = f_out(prt)
            if f_in:
                assert f_in(prt)[1]  # `pos` must be > 0 if match
            ret.append(prt)

        if clen:
            end = cidx + clen
            ret.append(pattern_out[cidx:end])

        return ''.join(ret)

    def _set_methods(self, methods, handler, meta=None):
        for meth in methods:
            self._methods[meth] = RouteMethod(self, meth, handler, meta)

    def set_method(self, method, handler, meta=None):
        if isinstance(method, str):
            method = [method]
        self._set_methods(method, handler, meta)

    def _raise_if_registered(self, method, candidate):
        registered = set(self._methods) & set(method)
        if registered:
            registered_fullnames = {
                m: self._methods[m].handler_fullname
                for m in registered
            }
            candidate_fullname = RouteMethod.get_func_fullname(candidate)
            raise RouteMethodError(
                f'Trying to register `{candidate_fullname}` '
                f'as handler for route methods `{self.rule}: {method}`, '
                f'but there are already registered: `{registered_fullnames}`'
            )

    def add_method(self, method, handler, meta=None):
        if isinstance(method, str):
            method = [method]
        self._raise_if_registered(method, handler)
        self._set_methods(method, handler, meta)

    def remove_method(self, method):
        if isinstance(method, str):
            method = [method]
        [self._methods.pop(m, None) for m in method]

    @property
    def methods(self):
        return dict(self._methods)

    def __getitem__(self, method):
        if isinstance(method, str):
            method = [method]
        for name in method:
            meth = self._methods.get(name)
            if meth:
                return meth
        raise RouteMethodError()

    def params_signature(self):
        return {name: [False, filter] for name, filter in zip(self.params, self.filters)}

    def __call__(self, method, *a, **kw):
        return self[method](*a, **kw)

    def __repr__(self):
        return '<%r {%s}>' % (
            self.rule, ', '.join([str(m) for m in self.methods.values()])
        )

    @classmethod
    def make_params_dict(cls, names, values):
        return {n: v for n, v in zip(names, values) if not n.startswith(cls.anon_prefix)}

    @classmethod
    def parse_rule(cls, rule):
        assert rule[0] == '/'
        rule = rule[1:]
        pattern = []
        pattern_out = []  # to build url (pattern without filter selectors)
        params = []
        filters = []
        filters_out = []
        anon_counter = 0
        for part, param, filter, filter_args, filter_selector in cls.parser.iter_parse(rule):
            filter_selector = filter_selector or ''
            if not part:  # it is param or/and filter
                part = '\r'
                if not param:
                    param = f'{cls.anon_prefix}{anon_counter}'
                    anon_counter += 1
                params.append(param)
                h, f_out = cls.make_filter(filter, filter_args)
                filters.append(h)
                filters_out.append(f_out)
            pattern.append(part + filter_selector)
            pattern_out.append(part)
        return ''.join(pattern), params, filters, ''.join(pattern_out), filters_out


class HookTypes(IntEnum):
    SIMPLE = 0
    PARTIAL = 1


class RouteKey(dict):
    @staticmethod
    def check_args(args):
        if len(args) > 1:
            raise TypeError('Expected `rule` or `pattern`, not both')

    def __init__(self, rule=None, *, pattern=None):
        self.check_args([a for a in [rule, pattern] if a is not None])
        if rule:
            super().__init__(rule=rule)
        else:
            super().__init__(pattern=pattern)


class RadiRouter:

    __slots__ = ('radidict', 'routes', 'hooks', 'named_routes')

    parse_rule = staticmethod(Route.parse_rule)

    def __init__(self):
        self.radidict = RadiDict()
        self.routes = dict()
        self.hooks = dict()
        self.named_routes = dict()

    @classmethod
    def to_pattern(cls, rule):
        return cls.parse_rule(rule)[0]

    def add(self, rule, methods, handler, name=None, *, meta=None, overwrite=False):
        if isinstance(methods, str):
            methods = [methods]
        methods = [_.upper() for _ in methods]
        return self._add(rule, methods, handler, name, meta=meta, overwrite=overwrite)

    def __getitem__(self, key):
        """Return route object by the name or rule or pattern.

        `key` treatment by type:
            `str` - name of the named route
            `set` - rule (must be single-item)
            `dict` - rule or pattern (must be single-item)

        """
        if isinstance(key, str):
            return self.named_routes.get(key)

        RouteKey.check_args(key)
        if isinstance(key, set):
            kwargs = dict(rule=[*key][0])
        elif isinstance(key, dict):
            kwargs = key.copy()
            if ('pattern' in kwargs):
                kwargs['route_pattern'] = kwargs.pop('pattern')
        else:
            raise TypeError('Item key must be instance of (str | set | dict)')
        return self._match(**kwargs)

    def remove(self, route = None, *, route_pattern = None, name = None):
        """
        use '*' to remove whole route-branch
        remove('/foo/bar/*') - remove all routes starts with `/foo/bar/`
        `/` - doesnt matter, so you can
        remove('/foo/ba*') - remove all routes starts with `/foo/ba`
        """
        if route is not None:
            if isinstance(route, str):
                route_pattern = self.to_pattern(route)
                route = None
            else:
                route_pattern = route.pattern
        elif name is not None:
            route = self.named_routes.pop(name)
            route_pattern = route.pattern

        self.radidict.remove(route_pattern)
        if route:
            del self.routes[route.pattern]
        else:
            if route_pattern.endswith('*'):
                route_pattern = route_pattern[:-1]
                patterns_to_del = [
                    pattern for pattern in self.routes
                    if pattern.startswith(route_pattern)
                ]
                [self.routes.pop(pattern) for pattern in patterns_to_del]
                self._remove_named_routers(set(patterns_to_del))
            else:
                self.routes.pop(route_pattern, None)
                self._remove_named_routers({route_pattern})

    def resolve(self, path, methods=None):
        """
        if `methods` passed, return a tuple of (end_point, err404_405):
                - end_point: [route_meth: RouteMethod, params: dict, hooks: list] | None
                - error: [code: int, reason: str, extra] | None
        else:
            return route: Route or None

        NOTE: ignores start/end slashes

        """
        path = path.strip('/')
        route, extra = self.radidict.get(path, allow_partial=True)

        if not methods:
            return route or None

        if not route:
            return None, [404, 'Not Found', extra]
        try:
            meth = route[methods]
            params = route.make_params_dict(extra['param_keys'], extra['param_values'])
            return [meth, params, extra['hooks']], None
        except RouteMethodError:
            allowed = ",".join(sorted(route.methods))
        return None, [405, "Method not allowed.", allowed]

    @staticmethod
    def hook_installer(route_hooks, hook, hook_type=HookTypes.SIMPLE):
        hook_type = HookTypes(hook_type)
        if not route_hooks:
            route_hooks = [None, None]
        route_hooks[hook_type] = hook
        return route_hooks

    def add_hook(self, rule, *args, **kwargs):
        route_pattern, params, filters, _, _ = self.parse_rule(rule)
        route_hooks = self._match(route_pattern = route_pattern, get_hooks = True)
        new_route_hooks = self.hook_installer(route_hooks, *args, **kwargs)
        if new_route_hooks is not None and new_route_hooks is not route_hooks:
            params = {name: [False, filter] for name, filter in zip(params, filters)}
            self.radidict.add_hooks(route_pattern, new_route_hooks, params)
            self.hooks[route_pattern] = new_route_hooks
        return route_pattern

    def get_hook(self, rule):
        route_pattern, params, filters, _, _ = self.parse_rule(rule)
        return self.hooks[route_pattern]

    def remove_hook(self, rule):
        route_pattern = self.to_pattern(rule)
        self.radidict.remove(route_pattern, hooks_only=True)
        self.hooks.pop(route_pattern, None)

    def _match(self, rule = None, filters = None, *, route_pattern = None, get_hooks = False):
        '''
        try to find the registered route, if filters is passed - performs filters comparision
        _match('/foo/:bar:re([^/]+)/baz') - will parse route and retrieve filters
        _match( 'foo/\r/baz', [make_filter('re', '[^/]+')] ) - the same as above
        _match(pattern, filters)  - is the same as -  _match(filters, route_pattern = pattern)
        _match(route_pattern = '/foo/\r/baz') - without filter comparision
            - used for removing, hook installation and general info
        '''

        if route_pattern is None:
            if filters is None:
                route_pattern, params, filters, _, _ = self.parse_rule(rule)
            else:
                # if `filters` is passed, but not `route_pattern`, then rule is pattern
                route_pattern = rule
        if route_pattern:
            assert route_pattern[0] != '/'
        node, mismatch, i, param_idx, stack = self.radidict._match(route_pattern, param_filters = filters)
        if not mismatch:
            return node[DATA] if not get_hooks else node[HOOKS]

    def _add(self, rule, methods, handler, name=None, *, meta=None, overwrite=False):
        route = Route(rule)
        route_ = self._match(route.pattern, route.filters)
        if route_:
            route = route_
        else:
            self.radidict.add(route.pattern, route, route.params_signature())
            self.routes[route.pattern] = route

        if overwrite:
            route.set_method(methods, handler, meta)
        else:
            route.add_method(methods, handler, meta)

        if name:
            registered = self.named_routes.get(name)
            if not overwrite and registered and registered is not route:
                raise RouteBuildError(f'Can`t register route, name `{name}` is already used')
            self.named_routes[name] = route
        return route

    def _remove_named_routers(self, pattern_set):
        for k, r in [*self.named_routes.items()]:
            if r.pattern in pattern_set:
                self.named_routes.pop(k)



import re

from .errors import RouteSyntaxError


def find_pos_after_close_paren(s, start=0, length = None):
    map_close = {
        '(': ')',
        '{': '}',
        '<': '>',
    }
    open_ = s[start]
    close_ = map_close[open_]
    L = len(s[start:]) if length is None else length
    i = start + 1
    done = False
    nested_level = 0
    while i < L:
        if s[i] == '\\':
            i += 2
            continue
        if s[i] == close_:
            if nested_level > 0:
                nested_level -= 1
            else:
                i += 1
                done = True
                break
        elif s[i] == open_:
            nested_level += 1
        i += 1
    if not done:
        return None
    return i


class SymStream:
    syntax_error = RouteSyntaxError

    def __init__(self, s: str):
        self.s = s
        self.L = len(s)
        self.pos = 0
        self.current = s[0] if s else None

    def _set_pos(self, pos):
        if not self.current:
            return
        if pos < self.L:
            self.pos = pos
            self.current = self.s[pos]
        else:
            self.pos = self.L
            self.current = None

    def next(self):
        if not self.current:
            return None
        self._set_pos(self.pos + 1)
        return self.current

    def eat(self, rex_str, group = None):
        if not self.current:
            return None
        i = self.pos
        s = self.s
        ret = None
        tmp = re.match(rex_str, s[i:])
        if tmp:
            j = i + tmp.end()
            ret = s[i:j] if group is None else tmp.group(group)
            self._set_pos(j)
        return ret

    def expect(self, rex_str, group = None):
        ret = self.eat(rex_str, group)
        if ret is None:
            raise self.syntax_error(
                f'Expected `{rex_str}` instead of `...{self.rest()}`\n\tin {self.s}'
            )
        return ret

    def expect_parenthesized(self):
        pos = find_pos_after_close_paren(self.s, self.pos, self.L)
        if pos is None:
            raise self.syntax_error(
                f'Missing closing parenthesis for `{self.current}` in `...{self.rest()}`\n\tin {self.s}'
            )
        ret = self.s[self.pos: pos]
        self._set_pos(pos)
        return ret

    def rest(self):
        return self.s[self.pos:]





class MixableMeta(type):
    __mixins_special__ = {'on_new', 'on_init'}

    def __new__(cls, name, bases, dct):
        mixins = []
        mixins_set = dct.get('_as_mixins')
        if mixins_set:
            bases_ = []
            for m in bases:
                if m in mixins_set:
                    mixins.append(m)
                else:
                    bases_.append(m)
            bases = tuple(bases_)
        if mixins:
            cls._mixin(dct, *mixins)
        out_cls = super().__new__(cls, name, bases, dct)
        return out_cls

    def __init__(cls, name, bases, dct):
        if name == 'Mixable':
            return

        cls_init = cls.__init__
        cls_new = cls.__new__

        def new_wrapper(cls, *a, **kw):
            self = cls_new(cls, *a, **kw)
            on_new_mixins = cls.__mixins_special__['on_new']
            [new(self, *a, **kw) for new in on_new_mixins]
            return self

        def init_wrapper(self, *a, **kw):
            cls_init(self, *a, **kw)
            on_init_mixins = self.__mixins_special__['on_init']
            [init(self, *a, **kw) for init in on_init_mixins]

        cls.__new__ = new_wrapper
        cls.__init__ = init_wrapper

    @classmethod
    def _mixin(cls, dct, *mixins):
        special = {k: [] for k in cls.__mixins_special__}
        slots = [*dct.get('__slots__', [])]
        for mixin_cls in mixins:
            mixin_slots = set(getattr(mixin_cls, '__slots__', []))
            for k, v in mixin_cls.__dict__.items():
                if k in mixin_slots:
                    continue
                if k == '__slots__':
                    slots.extend(v)
                elif k in special:
                    special[k].append(v)
                elif not k.startswith('__') and k not in dct:
                    dct[k] = v
        dct['__mixins_special__'] = special
        if slots:
            dct['__slots__'] = tuple(set(slots))


class Mixable(metaclass=MixableMeta):
    pass


if __name__ == '__main__':

    class BaseQQ:
        def __new__(cls, *a, **kw):
            self = super().__new__(cls)
            self.base = 'base'
            return self

    class qmixin:
        __slots__ = ('_a',)

        def on_init(self, *a, **kw):
            self._a = 'a'

        @classmethod
        def get_name(cls):
            return cls.__name__

        @property
        def a(self):
            return self._a

    class QQ(Mixable, BaseQQ, qmixin):
        __slots__ = ('b',)
        _as_mixins = {qmixin}

        def __init__(self):
            super().__init__()

        def __getattr__(self, k):
            pass

    assert QQ.get_name() == QQ.__name__

    qq = QQ()

    assert qq.a == 'a'
    assert qq.base == 'base'
